"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__

# NOTE when you do this PLY adds a module name self. to all tokens
# not sure if this is an unexpected feature or not. It can be worked around
# but for now direct import seems the safest
from . import lex
from . import yacc
import os
import math

try:
    from importlib import reload  # Python 3
except ImportError:
    pass


class MyInfixLexer:
    """
    This parser has been cranked to handle Python infix, numpy and MathML2.0 prefix expressions
    """

    debug = 0
    LexOK = True
    LexErrors = None
    __pwcntr__ = 0

    MathmlToNumpy_funcs = {
        'pow': 'pow',
        'root': 'pow',
        'abs': 'abs',
        'exp': 'math.exp',
        'ln': 'math.log',
        'log': 'math.log10',
        'floor': 'numpy.floor',
        'ceiling': 'numpy.ceil',
        'factorial': None,
        'sin': 'numpy.sin',
        'cos': 'numpy.cos',
        'tan': 'numpy.tan',
        'sec': None,
        'csc': None,
        'cot': None,
        'sinh': 'numpy.sinh',
        'cosh': 'numpy.cosh',
        'tanh': 'numpy.tanh',
        'sech': None,
        'csch': None,
        'coth': None,
        'arcsin': 'numpy.arcsin',
        'arccos': 'numpy.arccos',
        'arctan': 'numpy.arctan',
        'arcsec': None,
        'arccsc': None,
        'arccot': None,
        'arcsinh': 'numpy.arcsinh',
        'arccosh': 'numpy.arccosh',
        'arctanh': 'numpy.arctanh',
        'arcsech': None,
        'arccsch': None,
        'arccoth': None,
        'eq': 'operator.eq',
        'neq': 'operator.ne',
        'gt': 'operator.gt',
        'geq': 'operator.ge',
        'lt': 'operator.lt',
        'leq': 'operator.le',
        'ceil': 'numpy.ceil',
        'sqrt': 'math.sqrt',  # libsbml aliases
        'equal': 'operator.eq',
        'not_equal': 'operator.ne',  # numpy2numpy aliases
        'greater': 'operator.gt',
        'greater_equal': 'operator.ge',  # numpy2numpy aliases
        'less': 'operator.lt',
        'less_equal': 'operator.le',  # numpy2numpy aliases
        'ne': 'operator.ne',
        'ge': 'operator.ge',
        'le': 'operator.le',  # operator2operator
        'piecewise': 'self._piecewise_',
        '_piecewise_': 'self._piecewise_',
        'not': 'operator.not_',
        'not_': 'operator.not_',
    }

    MathmlToNumpy_symb = {
        'notanumber': 'numpy.NaN',
        'pi': 'numpy.pi',
        'infinity': 'numpy.Infinity',
        'exponentiale': 'numpy.e',
        'true': 'True',
        'false': 'False',
        'True': 'True',
        'False': 'False',
    }

    SymbolReplacements = None
    FunctionReplacements = None

    MathmlToInfix = {
        'and': 'and',
        'or': 'or',
        'true': 'True',
        'false': 'False',
        'xor': 'xor',
    }

    precedence = (
        ('left', 'ANDOR'),
        ('left', 'EQUISYMB'),
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIVIDE'),
        ('left', 'POWER'),
        ('right', 'UMINUS'),
    )

    # List of token names
    tokens = (
        'REAL',
        'INT',
        'PLUS',
        'MINUS',
        'TIMES',
        'DIVIDE',
        'POWER',
        'LPAREN',
        'RPAREN',
        'NOTEQUALS',
        'NAME',
        'ANDOR',
        'COMMA',
        'EQUISYMB',
        'PIECEWISE',
        'DELAY',
    )

    def __init__(self):
        self.LexErrors = []

        self.Int = r'\d+'  # Integer
        self.Dec = self.Int + '\.' + self.Int  # Decimal
        self.Exp = r'([E|e][\+|\-]?)' + self.Int  # Exponent
        self.Real = (
            self.Dec + '(' + self.Exp + ')?' + '|' + self.Int + self.Exp
        )  # Real - dec w/o optional exp or int with exp

        # Simple tokens
        self.t_REAL = self.Real
        self.t_INT = self.Int
        self.t_PLUS = r'\+'
        self.t_MINUS = r'-'
        self.t_TIMES = r'\*'
        self.t_DIVIDE = r'/'
        self.t_POWER = '\*\*'
        self.t_LPAREN = r'\('
        self.t_RPAREN = r'\)'
        self.t_COMMA = r','
        self.t_NOTEQUALS = r'!='

    def t_NAME(self, t):
        r'numpy\.[\w]*|math\.[\w]*|operator\.[\w]*|random\.[\w]*|self\.[\w]*|[a-zA-Z_][\w]*'

        # names are defined as anything starting with a letter OR numpy. math. or operator.
        t.type = 'NAME'
        # allow self. to be in names but always remove! dodgy testing stage
        t.value = t.value.replace('self.', '')
        if t.value == 'and':
            t.type = 'ANDOR'
            t.value = ' %s ' % t.value
        elif t.value == 'or':
            t.type = 'ANDOR'
            t.value = ' %s ' % t.value
        elif t.value == 'xor':
            t.type = 'ANDOR'
            t.value = ' %s ' % t.value
        elif t.value == 'piecewise':
            t.type = 'PIECEWISE'
            t.value = ' %s ' % t.value
        elif t.value == 'delay':
            t.type = 'DELAY'
            t.value = ' %s ' % t.value
        return t

    def t_EQUISYMB(self, t):
        r'>=|<=|!=|==|>|<'
        t.type = 'EQUISYMB'
        t.value = ' {} '.format(t.value)
        ##  'EQUISYMB', t.value
        return t

    # Define a rule so we can track line numbers
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    # A string containing ignored characters (spaces and tabs)
    t_ignore = ' \t'

    # Error handling rule
    def t_error(self, t):
        print("Illegal character '{}'".format(t.value[0]))
        self.LexErrors.append(t.value[0])
        self.LexOK = False
        t.lexer.skip(1)

    # Build the lexer
    def buildlexer(self, **kwargs):
        # try and find a temporary workspace
        if 'TMP' in os.environ:
            tempDir = os.environ['TMP']
        elif 'TEMP' in os.environ:
            tempDir = os.environ['TEMP']
        else:
            tempDir = os.getcwd()
        os.chdir(tempDir)
        self.lexer = lex.lex(object=self, **kwargs)

    # Test it output
    def testlexer(self, data):
        self.lexer.input(data)
        while 1:
            tok = self.lexer.token()
            if not tok:
                break
            print(tok)


class MyInfixParser(MyInfixLexer):
    ParseOK = True
    SymbolErrors = None
    ModelUsesNumpyFuncs = 0
    names = None
    functions = None
    output = None
    input = None
    name_prefix = '<pre>'
    name_suffix = '<suf>'
    _runCount = 0
    _runCountmax = 20
    __pwcntr__ = 0
    piecewises = None
    DelayRemoved = False

    def __init__(self):
        MyInfixLexer.__init__(self)
        self.ParseErrors = []
        self.names = []
        self.functions = []
        self.SymbolErrors = []
        self.piecewises = {}

    def setNameStr(self, prefix, suffix):
        self.name_prefix = str(prefix)
        self.name_suffix = str(suffix)

    def p_error(self, t):
        try:
            self.ParseErrors.append(t)
        except:
            print('p_error generated a parsing error')
        #fixes userwarning 183
        tok = self.parser.token()
        #tok = yacc.token()
        return tok

    def p_infix(self, t):
        '''Expression : Expression PLUS Expression
                      | Expression MINUS Expression
                      | Expression TIMES Expression
                      | Expression DIVIDE Expression
                      | Expression EQUISYMB Expression
                      | Expression ANDOR Expression
                      | Power
                      | Number
                      | Func
                      | Equivalence
                      | Piecewise
                      | NotEquals
                      | Delay'''
        # |UMINUS : add if the
        #  alternative for p_uminus is used

        if len(t.slice) == 4:
            t[0] = t[1] + t[2] + t[3]
        else:
            t[0] = t[1]

    def p_notequals(self, t):
        '''NotEquals : NOTEQUALS'''
        t[0] = t[1]

    def p_power(self, t):
        '''Power : Expression POWER Expression'''

        ##  t[0] = 'numpy.power('+ t[1] + ',' + t[3] + ')' #changed to make it DeriVar compatible
        t[0] = 'pow(' + t[1] + ',' + t[3] + ')'

    def p_number(self, t):
        '''Number : REAL
                  | INT
                  | NAME'''
        try:
            t[0] = str(float(t[1]))
        except ValueError:
            if t[1].strip() in self.MathmlToNumpy_symb:
                if self.MathmlToNumpy_symb[t[1]] == None:
                    self.SymbolErrors.append(t[1])
                    print('\nSymbol \"{}\" not yet supported by PySCeS.'.format(t[1]))
                    t[0] = 'unknown_symbol_' + t[1]
                else:
                    t[0] = self.MathmlToNumpy_symb[t[1]]
                self.ModelUsesNumpyFuncs = 1
            elif (
                t[1].replace('numpy.', '').replace('math.', '').replace('operator.', '')
                in self.MathmlToNumpy_symb
            ):
                t[0] = t[1]
            else:
                if (
                    self.SymbolReplacements != None
                    and t[1].strip() in self.SymbolReplacements
                ):
                    # replace symb --> prefix.replacement.suffix
                    if self.SymbolReplacements[t[1]] not in self.names:
                        self.names.append(self.SymbolReplacements[t[1]])
                    t[0] = (
                        self.name_prefix
                        + self.SymbolReplacements[t[1]]
                        + self.name_suffix
                    )
                elif (
                    self.FunctionReplacements != None
                    and t[1].strip() in self.FunctionReplacements
                ):
                    # replace symb --> (replacement)
                    t[0] = '({})'.format(self.FunctionReplacements[t[1]])
                else:
                    if t[1] not in self.names:
                        self.names.append(t[1])
                    t[0] = self.name_prefix + t[1] + self.name_suffix

    def p_uminus(self, t):
        '''Expression : MINUS Expression %prec UMINUS'''
        # Alternative '''UMINUS : MINUS Expression'''

        t[0] = t[1] + t[2]

    def p_equivalence(self, t):
        '''Equivalence : ANDOR LPAREN Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                        | ANDOR LPAREN Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression COMMA Expression RPAREN
                       '''

        # this is an almighty hack but i cant see any other way to do it right now ... ALL SUGGESTIONS WELCOME
        # changes and(a,b, .....) or(a,b, .....) to (a and b and ...) (a or b or ...)
        # and not(b) into self._not_(b)
        ##  print 'equivalence', len(t), t[:]
        t[1] = t[1].strip()
        if t[1] in self.MathmlToInfix:
            t[0] = t[2]
            for tt in range(3, len(t)):
                if t[tt] == ',':
                    if t[1] != 'xor':
                        t[0] += ' {} '.format(t[1])
                    else:
                        t[0] += ' {} '.format('!=')
                else:
                    t[0] += '{}'.format(t[tt])

    def p_piecewise(self, t):
        '''Piecewise : PIECEWISE LPAREN ArgListSemiCol RPAREN'''

        t[1] = t[1].strip()
        t[0] = self.MathmlToNumpy_funcs[t[1]] + t[2] + t[3] + t[4]
        pw = t[3].split(';')
        for p in range(len(pw)):
            pw[p] = pw[p].strip()
        name = '__pw{}__'.format(self.__pwcntr__)
        if len(pw) == 3:
            self.__pwcntr__ += 1
            self.piecewises.update({name: {0: [pw[1], pw[0]], 'other': pw[2]}})
        else:
            self.__pwcntr__ += 1
            self.piecewises.update({name: {}})
            if math.modf(len(pw) / 2.0)[0] != 0.0:
                self.piecewises[name].update({'other': pw.pop(-1)})
            else:
                self.piecewises[name].update({'other': None})
            for p in range(0, len(pw), 2):
                self.piecewises[name].update({p: [pw[p + 1], pw[p]]})
        t[0] = self.name_prefix + name + self.name_suffix

    def p_delay(self, t):
        '''Delay : DELAY LPAREN Expression COMMA Expression RPAREN'''

        # for now we just remove the delay on the expression
        # updated 201528 now requires a function delayFunc(var, delay) to deal with it
        self.DelayRemoved = True
        t[0] = '__delayFunc__({}, {})'.format(t[3], t[5])

    def p_function(self, t):
        '''Func : LPAREN ArgList RPAREN
                | NAME LPAREN ArgList RPAREN
                | NAME LPAREN RPAREN
                '''
        # this is to match NAME() which as far as I know is unique to object __calls__
        # as well as differentiate between bracketed functions and expressions:
        # func( S1 ) and ( S/S05 )

        if len(t) == 4:
            if t[1] == '(':
                t[0] = t[1] + t[2] + t[3]
            else:
                t[0] = self.name_prefix + t[1] + t[2] + t[3]
        # convert root(degree,<expr>) to pow(<expr>, 1/degree)
        elif t[1].strip() == 'root':
            t[1] = self.MathmlToNumpy_funcs[t[1]]
            t[3] = '{}, {}'.format(
                t[3][t[3].index(',') + 1 :], 1.0 / float(t[3][: t[3].index(',')]),
            )
            t[0] = t[1] + t[2] + t[3] + t[4]
        elif t[1].strip() in self.MathmlToNumpy_funcs:
            if self.MathmlToNumpy_funcs[t[1]] == None:
                self.SymbolErrors.append(t[1])
                print('\nFunction \"{}\" not supported by PySCeS'.format(t[1]))
                t[0] = 'unknown_function_' + t[1] + t[2] + t[3] + t[4]
            else:
                try:
                    t[0] = self.MathmlToNumpy_funcs[t[1]] + t[2] + t[3] + t[4]
                except Exception as EX:
                    print('Function Parse error 1 (please report!)\n', EX)
            self.ModelUsesNumpyFuncs = True
        else:
            # t[0] = t[1] + t[2] + t[3]
            # or a numpy fucntion
            if t[1][:6] == 'numpy.' or t[1][:5] == 'math.' or t[1][:9] == 'operator.':
                t[0] = t[1] + t[2] + t[3] + t[4]
            else:
                # assume some arbitrary function definition
                t[0] = self.name_prefix + t[1] + t[2] + t[3] + t[4]
                # add to list of functions
                if t[1] not in self.functions:
                    self.functions.append(t[1])

    # adapted from Andrew Dalke's GardenSnake
    # http://www.dalkescientific.com/writings/diary/GardenSnake.py
    # function arguments f(x,y,z)
    def p_arglist(self, t):
        '''ArgList : Expression
                   | ArgList COMMA Expression'''
        try:
            if len(t) == 2:
                t[0] = t[1]
            elif len(t) == 4:
                t[0] = t[1] + t[2] + t[3]
        except Exception as EX:
            print('Function ArgList error (please report!)\n', EX)

    # expression list f(g(x,y); g(a,b))
    def p_arglist_semicol(self, t):
        '''ArgListSemiCol : Expression
                   | ArgListSemiCol COMMA Expression'''
        try:
            if len(t) == 2:
                t[0] = t[1]
            elif len(t) == 4:
                t[0] = t[1] + '; ' + t[3]
        except Exception as EX:
            print('Function ArgList error (please report!)\n', EX)

    def buildparser(self, **kwargs):
        self.parser = yacc.yacc(module=self, **kwargs)

    def parse(self, data):
        self.ParseErrors = []
        self.LexErrors = []
        self.SymbolErrors = []
        self.names = []
        self.functions = []
        self.input = data
        self.ParseOK = True
        self.LexOK = True
        self.piecewises = {}
        self.DelayRemoved = False
        self.output = self.parser.parse(data)
        ##  assert len(self.SymbolErrors) == 0, '\nUndefined symbols:\n%s' % self.SymbolErrors
        ##  if len(self.SymbolErrors) != 0:
        ##  print '\nUndefined symbols:\n%s' % self.SymbolErrors
        assert self.LexOK, '\nLexer Failure:\n{}'.format(self.LexErrors)
        assert self.ParseOK, '\nParser Failure:\n{}'.format(self.ParseErrors)
        self._runCount += 1
        self.SymbolReplacements = None
        self.FunctionReplacements = None
        if self._runCount > self._runCountmax:
            self._runCount == 0
            # we're back !!!
            reload(lex)
            reload(yacc)



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = """
          PyscesScan
          ----------

          PySCeS classes for continuations and multi-dimensional parameter scans
          """

import scipy
from pysces.PyscesUtils import TimerBox

__psyco_active__ = 0


class Scanner(object):
    """
    Arbitrary dimension generic scanner. This class is initiated with
    a loaded PySCeS model and then allows the user to define scan parameters
    see self.addScanParameter() and user output see self.addUserOutput().
    Steady-state results are always stored in self.SteadyStateResults while
    user output can be found in self.UserOutputResults - brett 2007.
    """

    genOn = True
    quietRun = False
    _MODE_ = 'state'
    HAS_USER_OUTPUT = False
    HAS_STATE_OUTPUT = True
    nan_on_bad_state = True
    MSG_PRINT_INTERVAL = 500
    __AnalysisModes__ = ('state', 'elasticity', 'mca', 'stability', 'null')
    invalid_state_list = None
    invalid_state_list_idx = None

    def __init__(self, mod):
        try:
            N = mod.nmatrix
        except Exception as ex:
            print(
                '\nSCANNER: Please load a model <i.e. mod. doLoad()> before trying to use it.'
            )

        self.GenDict = {}
        self.GenOrder = []
        self.ScanSpace = []
        self.mod = mod
        self.SteadyStateResults = []
        self.UserOutputList = []
        self.UserOutputResults = []
        self.scanT = TimerBox()

    def testInputParameter(self, name):
        """
        This tests whether a str(name) is an attribute of the model
        """
        if hasattr(self.mod, name):
            return True
        else:
            return False

    def resetInputParameters(self):
        """
        Just remembered what this does, I think it resets the input model
        parameters after a scan run.
        """
        for key in self.GenDict:
            setattr(self.mod, key, self.GenDict[key][6])

    def addUserOutput(self, *kw):
        """
        Add output parameters to the scanner as a collection of one or more
        string arguments ('O1','O2','O3', 'On'). These are evaluated at
        each iteration of the scanner and stored in the self.UserOutputResults
        array. The list of output is stored in self.UserOutputList.
        """
        output = [attr for attr in kw if type(attr) == str]
        # print output
        self.HAS_USER_OUTPUT = True
        ModeList = list(self.__AnalysisModes__)
        MaxMode = 0
        for attr in output:
            if attr[:2] == 'ec' and self._MODE_ != 'null':
                cM = ModeList.index('elasticity')
                if cM > MaxMode:
                    self._MODE_ = 'elasticity'
                    MaxMode = cM
            elif attr[:2] == 'cc' and self._MODE_ != 'null':
                cM = ModeList.index('mca')
                if cM > MaxMode:
                    self._MODE_ = 'mca'
                    MaxMode = cM
            elif attr[:6] == 'lambda' and self._MODE_ != 'null':
                cM = ModeList.index('stability')
                if cM > MaxMode:
                    self._MODE_ = 'stability'
                    MaxMode = cM
        print('MaxMode', MaxMode)
        self.UserOutputList = output

    def addScanParameter(self, name, start, end, points, log=False, follower=False):
        """
        Add a parameter to scan (an axis if you like) input is:

        - str(name) = model parameter name
        - float(start) = lower bound of scan
        - float(end) = upper bound of scan
        - int(points) = number of points in scan range
        - bool(log) = Use a logarithmic (base10) range
        - bool(follower) = Scan parameters can be leaders i.e. an independent axis or a "follower" which moves synchronously with the previously defined parameter range.

        The first ScanParameter cannot be a follower.
        """
        offset = 1
        assert self.testInputParameter(name), (
            '\nSCANNER: Model does not have an attribute \"%s\" \n' % name
        )
        assert (
            name not in self.GenOrder
        ), '\nSCANNER: This operation is currently not allowed\n'
        if not len(self.GenOrder) == 0:
            for el in self.GenOrder:
                if self.GenDict[el][4] == False:  # test if not follower
                    offset = offset * self.GenDict[el][2]  # increment offset
            if follower == True:
                prevpar = self.GenOrder[-1]  # previous parameter, i.e. leader
                offset = self.GenDict[prevpar][5]  # don't increment for follower
                if points != self.GenDict[prevpar][2]:
                    print(
                        'SCANNER: Follower parameter needs to iterate over same number of\npoints as leader...resetting points.'
                    )
                    points = self.GenDict[prevpar][2]
        else:
            if follower == True:
                follower = False
                print(
                    'SCANNER: Inner range cannot be a follower ... resetting to leader'
                )
        self.GenDict.setdefault(
            name, (start, end, points, log, follower, offset, getattr(self.mod, name))
        )
        setattr(self, name + '_test', self.stepGen(offset))
        setattr(self, name, self.rangeGen(name, start, end, points, log))
        if name not in self.GenOrder:
            self.GenOrder.append(name)

    def makeRange(self, start, end, points, log):
        """
        Should be pretty self evident it defines a range:

        - float(start)
        - float(end)
        - int(points)
        - bool(log)
        """
        if log:
            rng = scipy.logspace(scipy.log10(start), scipy.log10(end), points)
        else:
            rng = scipy.linspace(start, end, points)
        return rng

    def rangeGen(self, name, start, end, points, log):
        """
        This is where things get more interesting. This function creates
        a cycling generator which loops over a parameter range.

        - *parameter* name
        - *start* value
        - *end* value
        - *points*
        - *log* scale
        """
        range = self.makeRange(start, end, points, log)
        cnt = 0
        while self.genOn:
            if next(getattr(self, name + '_test')):
                if cnt >= len(range) - 1:
                    cnt = 0
                else:
                    cnt += 1
            yield range[cnt]

    def stepGen(self, offset):
        """
        Another looping generator function. The idea here is to create
        a set of generators for the scan parameters. These generators then
        all fire together and determine whether the range generators should
        advance or not. Believe it or not this dynamically creates the matrix
        of parameter values to be evaluated.
        """
        val = 0
        while self.genOn:
            if val >= offset:
                val = 0
                yield True
            else:
                yield False
            val += 1

    def __nextStep__(self):
        """
        Fire all step generators
        """
        return [next(getattr(self, genN + '_test')) for genN in self.GenOrder]

    def __nextValue__(self):
        """
        Fire all range generators
        """
        return [next(getattr(self, genN)) for genN in self.GenOrder]

    def setModValue(self, name, value):
        """
        An easy one, assign value to name of the instantiated PySCeS model attribute
        """
        assert hasattr(self.mod, name), (
            '\nModel does not have an attribute: %s \n' % name
        )
        setattr(self.mod, name, float(value))

    def RunAgain(self):
        """
        While it is impossible to change the generator/range structure of a scanner
        (just build another one) you can 'in principle' change the User Output and run
        it again.
        """
        self.Run(ReRun=True)

    def Run(self, ReRun=False):
        """
        Run the parameter scan
        """
        self.scanT.normal_timer('RUN')
        self._MODE_ = self._MODE_.lower()
        assert self._MODE_ in self.__AnalysisModes__, (
            '\nSCANNER: \"%s\" is not a valid analysis mode!' % self._MODE_
        )
        if self.quietRun:
            self.mod.SetQuiet()
        else:
            self.mod.SetLoud()
        if ReRun:
            self.ScanSpace = []
            self.UserOutputResults = []
            self.SteadyStateResults = []
        self.invalid_state_list = []
        self.invalid_state_list_idx = []

        if self.nan_on_bad_state:
            self.mod.__settings__["mode_state_nan_on_fail"] = True
            # self.mod.mode_state_nan_on_fail = True
        Tsteps = 1
        for gen in self.GenOrder:
            if self.GenDict[gen][4] == False:  # don't increase Tsteps for follower
                Tsteps *= self.GenDict[gen][2]
        print(next(self.scanT.RUN))
        analysis_counter = 0
        print('SCANNER: Tsteps', Tsteps)
        pcntr = 0
        for step in range(Tsteps):
            pars = self.__nextValue__()
            # if pars not in self.ScanSpace:
            self.ScanSpace.append(pars)
            for par in range(len(self.GenOrder)):
                self.setModValue(self.GenOrder[par], pars[par])
            analysis_counter += 1
            self.Analyze()
            if not self.mod.__StateOK__:
                self.invalid_state_list.append(pars)
                self.invalid_state_list_idx.append(analysis_counter)
            self.StoreData()
            pcntr += 1
            if pcntr >= self.MSG_PRINT_INTERVAL:
                print('\t', analysis_counter, next(self.scanT.RUN))
                pcntr = 0
        self.ScanSpace = scipy.array(self.ScanSpace)
        self.SteadyStateResults = scipy.array(self.SteadyStateResults)
        self.UserOutputResults = scipy.array(self.UserOutputResults)
        self.resetInputParameters()
        if self.nan_on_bad_state:
            self.mod.mode_state_nan_on_fail = False
        print("\nSCANNER: %s states analysed\n" % analysis_counter)
        if len(self.invalid_state_list) > 0:
            print('Bad steady states encountered at:\n')
            print(self.invalid_state_list)
            print(self.invalid_state_list_idx)

    def Analyze(self):
        """
        The analysis method, the mode is automatically set by the self.addUserOutput()
        method but can be reset by the user.
        """
        if self._MODE_ == 'state':
            self.mod.doState()
        elif self._MODE_ == 'elasticity':
            self.mod.doElas()
        elif self._MODE_ == 'mca':
            self.mod.doMca()
        elif self._MODE_ == 'stability':
            self.mod.doEigenMca()
        elif self._MODE_ == 'null':
            self.HAS_STATE_OUTPUT = False
        self.mod.User_Function()

    def StoreData(self):
        """
        Internal function which concatenates and stores the data generated by Analyze.
        """
        if self.HAS_STATE_OUTPUT and self._MODE_ != 'null':
            self.SteadyStateResults.append(
                scipy.hstack((self.mod.state_species, self.mod.state_flux))
            )
        if self.HAS_USER_OUTPUT and self._MODE_ != 'null':
            self.UserOutputResults.append(
                scipy.array([getattr(self.mod, res) for res in self.UserOutputList])
            )

    def getOutput(self):
        """
        Will be the new output function.
        """
        pass

    def getResultMatrix(self, stst=False, lbls=False):
        """
        Returns an array of result data. I'm keepin this for backwards compatibility but
        it will be replaced by a getOutput() method when this scanner is updated to use
        the new data_scan object.

        - *stst* add steady-state data to output array
        - *lbls* return a tuple of (array, column_header_list)

        If *stst* is True output has dimensions [scan_parameters]+[state_species+state_flux]+[Useroutput]
        otherwise [scan_parameters]+[Useroutput].
        """
        output_array = None
        labels = []
        if stst:
            if self.HAS_USER_OUTPUT:
                output_array = scipy.hstack(
                    [self.ScanSpace, self.SteadyStateResults, self.UserOutputResults]
                )
                labels = (
                    self.GenOrder
                    + list(self.mod.species)
                    + list(self.mod.reactions)
                    + self.UserOutputList
                )
            else:
                output_array = scipy.hstack([self.ScanSpace, self.SteadyStateResults])
                labels = (
                    self.GenOrder + list(self.mod.species) + list(self.mod.reactions)
                )
        else:
            output_array = scipy.hstack([self.ScanSpace, self.UserOutputResults])
            labels = self.GenOrder + self.UserOutputList
        if lbls:
            return output_array, labels
        else:
            return output_array


class PITCONScanUtils(object):
    """
    Static Bifurcation Scanning utilities using PITCON, call with loaded model object.
    Hopefully nobody else was trying to use the older class as it was horrible. This new
    one is is leaner, meaner and pretty cool ;-)
    """

    model = None
    timer = None
    pitcon_range_low = None
    pitcon_range_high = None
    pitcon_scan_density = None
    pitcon_scan_parameter = None
    pitcon_scan_parameter_3d = None
    pitcon_res = None

    user_output = None
    res_user = None

    res_idx = None
    res_metab = None
    res_flux = None
    res_eigen = None

    def __init__(self, model):
        """
        Instantiate PITCONScanUtils with a loaded PySCeS model instance.
        """
        self.model = model
        self.timer = TimerBox()

    def setUserOuput(self, *args):
        """
        Set the user output required as n string arguments.
        """
        self.user_output = [str(arg) for arg in args]

    def runContinuation(
        self, parameter, low, high, density, par3d=None, logrange=True, runQuiet=True
    ):
        """
        Run the continuation using the following parameters:

        Args:

        - parameter = str(the parameter to be scanned)
        - low = float(lower bound)
        - high = float(upper bound)
        - density = int(the number of initial points)
        - par3d = float(extra 3d parameter to insert into the output array) this parameter is not set ONLY used in output
        - logrange = boolean [default = True], if True generate the result using logspace(log10(low), log10(high), density) otherwise use a linear range
        - runQuiet = boolean [default = True], if True do not display intermediate results to screen, disable for debugging

        After running the continuation the results are stored in numpy arrays

        - mod.res_idx  = scan parameter values (and optionally par3d)
        - mod.res_metab = steady-state species concentrations
        - mod.res_flux = steady-state flux values

        """

        self.pitcon_scan_density = density
        self.pitcon_scan_parameter = parameter
        self.pitcon_scan_parameter_3d = par3d
        if logrange:
            self.pitcon_range_low = scipy.log10(low)
            self.pitcon_range_high = scipy.log10(high)
            self.model.pitcon_par_space = scipy.logspace(
                self.pitcon_range_low, self.pitcon_range_high, self.pitcon_scan_density
            )
        else:
            self.pitcon_range_low = low
            self.pitcon_range_high = high
            self.model.pitcon_par_space = scipy.linspace(
                self.pitcon_range_low, self.pitcon_range_high, self.pitcon_scan_density
            )

        self.model.pitcon_flux_gen = 1
        if runQuiet:
            self.model.SetQuiet()
        else:
            self.model.SetLoud()

        if self.pitcon_scan_parameter_3d != None:
            self.pitcon_res = self.model.PITCON(
                self.pitcon_scan_parameter, self.pitcon_scan_parameter_3d
            )
            self.res_idx = self.pitcon_res[:, :2]
            self.res_metab = self.pitcon_res[:, 2 : len(self.model.species) + 2 :]
            self.res_flux = self.pitcon_res[:, len(self.model.species) + 2 :]
        else:
            self.pitcon_res = self.model.PITCON(self.pitcon_scan_parameter)
            self.res_idx = self.pitcon_res[:, 0]
            self.res_idx = self.res_idx.reshape(self.res_idx.shape[0], 1)
            self.res_metab = self.pitcon_res[:, 1 : len(self.model.species) + 1]
            self.res_flux = self.pitcon_res[:, len(self.model.species) + 1 :]

        print('\n\tContinuation complete\n')

    def analyseData(self, analysis='elas'):
        """
        Performs "analysis" on the PITCON generated set of steady-state results where analysis is:

        - 'elasv' = variable elasticities
        - 'elasp' = parameter elasticities
        - 'elas'  = all elasticities
        - 'mca'   = control coefficients
        - 'resp'  = response coefficients
        - 'eigen' = eigen values
        - 'all'   = all of the above

        Higher level analysis types automatically enable the lower level analysis needed e.g.
        selecting 'mca' implies 'elasv' etc. User output defined with mod.setUserOutput() is
        stored in the mod.res_user array.
        """
        # reset result lists
        self.res_eigen = []
        self.res_user = []

        step_cntr = 0
        self.timer.step_timer('st1', self.res_idx.shape[0])
        for row in range(self.res_idx.shape[0]):
            self.timer.st1_cstep += 1
            step_cntr += 1
            if self.timer.st1_cstep == 1 or step_cntr == 200:
                print('\n', next(self.timer.st1))
                step_cntr = 0

            PitParVal = self.res_idx[row, -1]
            setattr(self.model, self.pitcon_scan_parameter, PitParVal)

            self.model.state_species = self.res_metab[row, :]
            self.model.state_flux = self.res_flux[row, :]
            self.model.SetStateSymb(self.model.state_flux, self.model.state_species)
            self.model.Forcing_Function()

            if analysis in ['elasv', 'elas', 'mca', 'resp', 'eigen', 'all']:
                self.model.EvalEvar(self.model.state_species, self.model.state_flux)
            if analysis in ['elasp', 'elas', 'resp', 'all']:
                self.model.EvalEpar(self.model.state_species, self.model.state_flux)
            if analysis in ['mca', 'resp', 'all']:
                self.model.EvalCC()
            if analysis in ['resp', 'all']:
                self.model.EvalRC()
            if analysis in ['eigen', 'all']:
                self.model.EvalEigen()
                self.res_eigen.append(self.model.eigen_values)
            state_out = []
            if self.user_output != None and len(self.user_output) > 0:
                for attr in self.user_output:
                    try:
                        state_out.append(getattr(self.model, attr))
                    except Exception as ex:
                        print(ex)
                self.res_user.append(state_out)
        self.res_user = scipy.array(self.res_user)
        print('\n\t%s evaluation complete\n' % analysis)

    def getArrayListAsArray(self, array_list):
        """
        Stack (concatenate) the list of arrays into a single array.
        """
        output = None
        if len(array_list) > 1:
            output = scipy.vstack(array_list)
        elif len(array_list) == 1:
            output = array_list[0]
        return output



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

# Init file used for distutils install



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

# begin header
import pysces
import numpy, scipy

from .KrakenNET import os, time, socket
from .KrakenNET import StatusServer, BasicServer
from .KrakenNET import HOSTNAME, BLOCK_SIZE, STATUS_PORT, PYSCES_PORT, CFSERVE_PORT

print('I AM:', HOSTNAME, 'using BLOCK_SIZE: %s' % BLOCK_SIZE)

# end header

__psyco_active__ = 0

from pysces import TimerBox


class PyscesServer(BasicServer):
    COMMAND_LIST = (
        'P_INIT',
        'P_LOAD',
        'P_SET_STR',
        'P_SET_FLOAT',
        'P_STATE',
        'P_SIM',
        'P_ELAS',
        'P_MCA',
        'P_SCAN',
        'P_CONTINUATION',
        'P_CONTINUATION_WITH_EIGEN',
        'P_SET_SETTINGS_STRING',
        'P_SET_SETTINGS_FLOAT',
        'P_CONTINUATION_WITH_EIGEN_USER',
        'P_CONTINUATION2D_WITH_EIGEN_USER',
    )
    model_name = None
    model = None

    def __init__(self, port, block_size, status_server=None, myname=None):
        BasicServer.__init__(self, port, block_size, status_server, myname)

    def P_INIT(self, *args):
        args = args[0]
        print('Args', args)
        self.RESULT = None
        self.setStatus('INITIALISING')
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.connect((self.client_address[0], CFSERVE_PORT))
        client.send('GET')
        GO = True
        model = ''
        while GO:
            data = client.recv(self.block_size)
            model += data
            if data == '':
                GO = False
        self.model_name = args[0]
        self.model = pysces.model(self.model_name, loader="string", fString=model)
        if self.debug:
            print('P_INIT ... ok.')
        self.setStatus('DONE_INITIALISE')
        return True

    def P_LOAD(self, *args):
        print('Loading:', self.model_name)
        self.RESULT = None
        self.model.doLoad()
        self.model.SetQuiet()
        if self.debug:
            print('P_LOAD ... ok.')
        self.setStatus('DONE_LOAD')
        return True

    def P_SET_SETTINGS_FLOAT(self, *args):
        args = args[0]
        print('Setting:', args)
        self.RESULT = None
        self.model.__settings__.update({args[0]: float(args[1])})
        ##  setattr(self.model, args[0], args[1])
        print(self.model.__settings__[args[0]])
        return True

    def P_SET_SETTINGS_STRING(self, *args):
        args = args[0]
        print('Setting:', args)
        self.RESULT = None
        self.model.__settings__.update({args[0]: args[1]})
        ##  setattr(self.model, args[0], args[1])
        print(self.model.__settings__[args[0]])
        return True

    def P_SET_STR(self, *args):
        args = args[0]
        print('Setting:', args)
        self.RESULT = None
        setattr(self.model, args[0], args[1])
        print(getattr(self.model, args[0]))
        return True

    def P_SET_FLOAT(self, *args):
        # note to self in future investigate numpy.float64 for this
        args = args[0]
        print('Setting:', args)
        self.RESULT = None
        setattr(self.model, args[0], float(args[1].strip()))
        print(getattr(self.model, args[0]))
        return True

    def P_STATE(self, *args):
        self.model.doState()
        self.RESULT = (self.model.state_species, self.model.state_flux)
        if self.debug:
            print('P_STATE ... ok.')
        self.setStatus('DONE_STATE')
        return True

    def P_ELAS(self, *args):
        self.model.doElas()
        self.RESULT = (self.model.elas_var, self.model.elas_par)
        if self.debug:
            print('P_ELAS ... ok.')
        self.setStatus('DONE_ELAS')
        return True

    def P_MCA(self, *args):
        self.model.doMca()
        self.RESULT = (self.model.mca_ci, self.model.mca_cjd, self.model.mca_csd)
        if self.debug:
            print('P_MCA ... ok.')
        self.setStatus('DONE_MCA')
        return True

    def P_SIM(self, *args):
        """Args: (float(time_end), int(points))"""
        self.setStatus('SIMULATING')
        args = args[0]
        print('Args', args)
        self.model.doSim(float(args[0]), float(args[1]))
        self.RESULT = self.model.sim_res
        if self.debug:
            print('P_STATE ... ok.')
        self.setStatus('DONE_SIM')
        return True

    def P_SCAN(self, *args):
        """Args: (par, out, par*[str(p),float(start),float(end),int(points),
                 log=(int(0/1)),follower=(int(0/1))], [out*str(o)])
        """
        self.setStatus('SCANNING')
        args = args[0]
        print('Args', args)
        commands = []
        cpoint = 1
        shift = 0
        for set in range(int(args[0])):
            ctmp = []
            for t in range(2, 8):
                cpoint += 1
                ctmp.append(args[t + shift].strip())
            commands.append(ctmp)
            shift += 6
        output = []
        for out in range(int(args[1])):
            cpoint += 1
            output.append(args[cpoint].strip())

        print('\n', commands)
        print(output, '\n')

        scan = pysces.Scanner(self.model)
        scan.quietRun = True
        for gen in commands:
            print(gen)
            print(
                gen[0],
                float(gen[1]),
                float(gen[2]),
                int(gen[3]),
                bool(int(gen[4])),
                bool(int(gen[5])),
            )

            scan.addGenerator(
                gen[0],
                float(gen[1]),
                float(gen[2]),
                int(gen[3]),
                log=bool(int(gen[4])),
                follower=bool(int(gen[5])),
            )
        scan.addUserOutput(*output)
        scan.Run()
        self.RESULT = scan.UserOutputResults
        self.setStatus('DONE_SCAN')
        del scan
        return True

    def P_CONTINUATION(self, *args):
        """
        [str(p), float(min), float(max), int(points), *float(y-value)]
        """
        self.setStatus('CONTINUING')
        args = args[0]
        print('Args', args)

        self.model.pitcon_par_space = scipy.logspace(
            scipy.log10(float(args[1].strip())),
            scipy.log10(float(args[2].strip())),
            int(args[3].strip()),
        )

        if len(args) == 4:
            self.RESULT = self.model.PITCON(args[0].strip())
        elif len(args) == 5:
            self.RESULT = self.model.PITCON(args[0].strip(), float(args[4].strip()))
        return True

    def P_CONTINUATION_WITH_EIGEN(self, *args):
        """
        [str(p), float(min), float(max), int(points), *float(y-value)]
        """
        self.setStatus('CONTINUING')
        args = args[0]
        print('Args', args)

        pscan = pysces.PITCONScanUtils(self.model)
        if len(args) == 4:
            pscan.runContinuation(
                args[0].strip(),
                float(args[1].strip()),
                float(args[2].strip()),
                int(args[3].strip()),
            )
        elif len(args) == 5:
            pscan.runContinuation(
                args[0].strip(),
                float(args[1].strip()),
                float(args[2].strip()),
                int(args[3].strip()),
                float(args[4].strip()),
            )
        pscan.analyseData(analysis='eigen')
        self.RESULT = (
            pscan.res_idx,
            pscan.res_metab,
            pscan.res_flux,
            pscan.getArrayListAsArray(pscan.res_eigen),
        )
        del pscan
        return True

    def P_CONTINUATION_WITH_EIGEN_USER(self, *args):
        """
        [str(p), float(min), float(max), int(points), float(y-value)]
        """
        self.setStatus('CONTINUING')
        args = args[0]
        print('Args', args)

        pscan = pysces.PITCONScanUtils(self.model)
        # TODO HARD CODED FOR NOW BUT NEEDS GENERALISATION
        pscan.setUserOuput('ecR1_A', 'ecR2_A')
        if len(args) == 4:
            pscan.runContinuation(
                args[0].strip(),
                float(args[1].strip()),
                float(args[2].strip()),
                int(args[3].strip()),
            )
        elif len(args) == 5:
            pscan.runContinuation(
                args[0].strip(),
                float(args[1].strip()),
                float(args[2].strip()),
                int(args[3].strip()),
                float(args[4].strip()),
            )
        pscan.analyseData(analysis='eigen')
        self.RESULT = (
            pscan.res_idx,
            pscan.res_metab,
            pscan.res_flux,
            pscan.getArrayListAsArray(pscan.res_eigen),
            pscan.res_user,
        )
        del pscan
        return True

    def P_CONTINUATION2D_WITH_EIGEN_USER(self, *args):
        """
        [str(p), float(min), float(max), int(points), float(y-value), str(userNx)]
        """
        self.setStatus('CONTINUING')
        args = args[0]
        print('Args', args)
        userout = []
        for a in range(len(args)):
            args[a] = args[a].strip()
            if a > 4:
                userout.append(args[a])

        pscan = pysces.PITCONScanUtils(self.model)
        if len(userout) > 0:
            pscan.setUserOuput(*userout)
        pscan.runContinuation(
            args[0], float(args[1]), float(args[2]), int(args[3]), float(args[4])
        )
        pscan.analyseData(analysis='all')
        self.RESULT = (
            pscan.res_idx,
            pscan.res_metab,
            pscan.res_flux,
            pscan.getArrayListAsArray(pscan.res_eigen),
            pscan.res_user,
        )
        del pscan
        return True


def StartPysCeSKrakenServer(
    STATUS_PORT=STATUS_PORT, PYSCES_PORT=PYSCES_PORT, BLOCK_SIZE=BLOCK_SIZE
):
    STATUSserve = StatusServer(STATUS_PORT, BLOCK_SIZE, 'KrakenStatus')
    STATUSserve.start()
    PYSCESserve = PyscesServer(PYSCES_PORT, BLOCK_SIZE, STATUSserve, 'KrakenPySCeS')
    PYSCESserve.start()
    PYSCESserve.join()


if __name__ == '__main__':
    ##  mod = pysces.model('isola_thesis')
    ##  mod.doLoad()
    ##  pscan = pysces.PITCONScanUtils(mod)
    ##  pscan.do2Dcontinuation('P',0.01,3000,5,par3d=0.1)
    ##  pscan.doEigenAnalysis()
    ##  print 'PITCON', pscan.pitcon_res[0]
    ##  print 'FLUX',   pscan.flux_res[0]
    ##  print 'EIGEN',  pscan.eigen_res[0]
    StartPysCeSKrakenServer()
    print('\nAll threads terminated ... exiting now.\n')



'''
This __init__ file is for a PySCeS "eXtension module" (pysx), the following variables can be set for each module
'''
pysx_name = 'Pysces module template'
pysx_oscompat = ['posix','nt']
pysx_base_class = 'CONTRIB_demo'

pysx_author = 'Brett G. Olivier'
pysx_email = 'bgoli@users.sourceforge.net'
pysx_affiliation = 'Triple-J Group for Molecular Cell Physiology'
pysx_web = 'http://pysces.sourceforge.net'
pysx_notes = ''

from .demotest import *



from Kraken import *
from KrakenDataTools import DataTools

os.chdir('c:/mypysces/btk')
##  os.chdir('/home/bgoli/mypysces/kraken')

START = time.time()
"""
Build controller and initialise active tentacles
Get a name for this task (task == collection of jobs)
"""
kcntrl = KrakenScanController()
kcntrl.setEnv(task_id='isola_cont_eigen', working_dir='c:\\mypysces\\btk',\
        model_file='2006_isola1_dem.psc', model_dir='c:\\mypysces\\pscmodels\\isola06')
kcntrl.startController()
"""
Set up a list of 'static' initialisation commands that will be used to
initialise each server before a job is sent to it
"""
I_list = ['P_INIT,' + kcntrl.kc_model_server.model_file_name, 'P_LOAD',\
    'P_SET_FLOAT,mode_state_init,3', 'P_SET_FLOAT,nleq2_nitmax,30',\
    'P_SET_FLOAT,mode_state_init3_factor,0.5', 'P_SET_FLOAT,pitcon_iter,20',\
    'P_SET_FLOAT,V2,100.0', 'P_SET_FLOAT,V3,100.0', 'P_SET_FLOAT,S05,1.0']
    ##  'P_SET_FLOAT,V2,1000.0', 'P_SET_FLOAT,V3,1000.0', 'P_SET_FLOAT,S05,1.0']
kcntrl.setInitCmds(I_list)


"""
Set up our continuation:
We want a parameter to be set with every new job/server and make use
of the deltaCommand() function. Here we use it to set our y-axis (A05) values
"""
# node split on yrange
# A05
yrange = 57
# P start points
xpoints = 200

# create a list of values A05
##  delta_list_vl = scipy.logspace(scipy.log10(0.1),scipy.log10(5.0), yrange)
##  delta_list_l = scipy.logspace(scipy.log10(5.1),scipy.log10(100.0), yrange)
##  delta_list_h = scipy.logspace(scipy.log10(100.1),scipy.log10(5000.0), yrange*3)
delta_list = scipy.logspace(scipy.log10(0.5),scipy.log10(1000.0), yrange)

# initiate an 'looping' generator with it
delta_gen = kcntrl.buildCycler(delta_list)

def deltaCommand():
    """
    This is used to setup conditions for each next-job-on-server
    Used in sendCmdListToOne() so may be one or more commands returned
    as a list
    """
    delta = 'P_SET_FLOAT,A05,%s' % next(delta_gen)
    # must return a list !!!
    return [delta]

kcntrl.deltaCommand = deltaCommand

"""
Set up our list of jobs, in this case a single continuation+eigen with
density xpoints at each of the delta (A05 values)
"""
J_list = []
for d in delta_list:
    ##  J_list.append('P_CONTINUATION,P,0.001,4.0e2,%s,%s' % (xpoints,d))
    J_list.append('P_CONTINUATION_WITH_EIGEN,P,0.01,3.8e4,%s,%s' % (xpoints,d))

kcntrl.setJobList(J_list)
kcntrl.Run(raw_data_dump=False)


DaResults = kcntrl.getAndClearResultList()

# Have data will travel

kdata = DataTools()

iRes = []
sRes = []
jRes = []
eRes = []
for res in DaResults:
    iRes.append(res[0])
    sRes.append(res[1])
    jRes.append(res[2])
    eRes.append(res[3])
del DaResults

def concatenateArrays(array_list):
    output = None
    for arr in range(len(array_list)):
        if type(array_list[arr]) == numpy.ndarray:
            if arr == 0:
                output = array_list[arr]
            else:
                output = scipy.concatenate((output,array_list[arr]))
    return output

FinalIArray = concatenateArrays(iRes)
print('Indexes[0]', FinalIArray[0])
FinalSArray = concatenateArrays(sRes)
FinalJArray = concatenateArrays(jRes)
FinalEArray = concatenateArrays(eRes)
print('FinalIArray', FinalIArray.shape, FinalIArray.dtype)
print('FinalSArray', FinalSArray.shape, FinalSArray.dtype)
print('FinalJArray', FinalJArray.shape, FinalJArray.dtype)
print(FinalEArray.shape, FinalEArray.dtype)
del iRes
del sRes
del jRes
del eRes

kcntrl.Dump((FinalIArray,FinalSArray,FinalJArray,FinalEArray), kcntrl.task_id+'_result_array.bin')

# get rid of negative values in axes and fluxes
FinalArrayCleanJ1 = []
FinalArrayCleanJ4 = []
FinalVTK = []
for r in range(FinalSArray.shape[0]):
    if FinalIArray[r,:].all() > 0.0 and FinalSArray[r,:].all() > 0.0 and \
        FinalJArray[r,:].all() > 0.0 and FinalIArray[r,1] >= 0.2 and FinalJArray[r,1] >= 0.08 and FinalIArray[r,1] <= 40000.0:
        FinalArrayCleanJ1.append([FinalIArray[r,0], FinalIArray[r,1], FinalJArray[r,0], max(FinalEArray[r,:].real)])
        FinalArrayCleanJ4.append([FinalIArray[r,0], FinalIArray[r,1], FinalJArray[r,3], max(FinalEArray[r,:].real)])
    else:
        FinalIArray[r,:] = 0.0
        FinalSArray[r,:] = 0.0
        FinalJArray[r,:] = 0.0
        FinalEArray[r,:] = 0.0
FinalArrayCleanJ1 = numpy.array(FinalArrayCleanJ1)
FinalArrayCleanJ4 = numpy.array(FinalArrayCleanJ4)

kdata.writeGnuPlot_3D(FinalArrayCleanJ1, kcntrl.task_id+'_J1')

def ScalarFunc(val):
    """
    Manipulate the scalar (z-axis) values
    Accepts and returns a double
    """
    if val >= 0.0:
        return 60.0
    else:
        return val


kdata.vtk_scalar_func = ScalarFunc

# prepare vtk data
vtkcoords = numpy.transpose(numpy.array((FinalArrayCleanJ1[:,0],FinalArrayCleanJ1[:,1],FinalArrayCleanJ1[:,2],FinalArrayCleanJ1[:,3]),'d'))
for row in range(vtkcoords.shape[0]):
    vtkcoords[row,:3] = scipy.log10(vtkcoords[row,:3])

print(vtkcoords[:5,:])
vtkcoords = kdata.GridSortLR(vtkcoords)
print(vtkcoords[:5,:])

kdata.writeVTK_UnstructuredGrid(vtkcoords, kcntrl.task_id+'_result_ug')



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import numpy, scipy, itertools, time
try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

from .KrakenNET import socket, time, cPickle
from .KrakenNET import SimpleClient, SimpleMultiReadClient, ThreadedClient
from .KrakenNET import ModelFileServer, TentacleScanner, ServerListLoader
from .KrakenNET import ServerStatusCheck, HOSTNAME, BLOCK_SIZE, PICKLE_PROTOCOL, STATUS_PORT, PYSCES_PORT, CFSERVE_PORT

print('I AM:', HOSTNAME)
print('Using BLOCK_SIZE: %s\n' % BLOCK_SIZE)

class KrakenController:
    kc_status_port = None
    kc_pysces_port = None
    kc_block_size = None

    kc_file_name = 'server_list'
    kc_directory_name = None
    kc_available_server_list = None
    kc_busy_server_list = None
    kc_tentacle_response = None
    kc_model_server = None
    kc_tentacle_scanner = None
    kc_tentacle_monitor = None
    kc_tentacle_process_map = None

    def __init__(self,status_port=STATUS_PORT, pysces_port=PYSCES_PORT, block_size=BLOCK_SIZE):
        self.kc_status_port = status_port
        self.kc_pysces_port = pysces_port
        self.kc_block_size = block_size
        self.kc_available_server_list = []
        self.kc_busy_server_list = []
        self.kc_tentacle_response = {}
        self.kc_tentacle_process_map = {}

        print(self.kc_status_port, self.kc_pysces_port, self.kc_block_size)

    def initialiseTentacles(self,file_name=None, directory_name=os.getcwd()):
        if file_name != None:
            self.kc_file_name = file_name
        self.kc_available_server_list = ServerListLoader().ReadFile(self.kc_file_name, directory_name)
        self.kc_tentacle_scanner = TentacleScanner(self.kc_available_server_list)
        print('Server list:')
        print(self.kc_available_server_list, '\n')
        assert len(self.kc_available_server_list) > 0, '\n* No servers in server list file'

    def getActiveTentacles(self):
        self.kc_tentacle_scanner.scan()
        self.kc_available_server_list = self.kc_tentacle_scanner.servers_ready
        self.kc_busy_server_list = self.kc_tentacle_scanner.servers_busy
        print('Server list:')
        print(self.kc_available_server_list, '\n')
        assert len(self.kc_available_server_list) > 0, '\n* No active servers'
        for t in self.kc_available_server_list:
            self.kc_tentacle_response.setdefault(t)
            self.kc_tentacle_process_map.setdefault(t)
            # maybe wil be needed later for reruns of getActiveTentacles
            ##  if self.kc_tentacle_process_map.has_key[t]:
                ##  dt = self.kc_tentacle_process_map.pop[t]
                ##  del dt

        ##  print self.kc_tentacle_response
        ##  print self.kc_tentacle_process_map


    def Run(self):
        pass

    def sendCmdListToAll(self, command_list, pause=0.5):
        assert type(command_list) == list or type(command_list) == tuple, '\n* Input must be a list or tuple'
        for server in self.kc_available_server_list:
            self.sendCmdListToOne(command_list, server)
            time.sleep(pause)

    def sendCmdListToOne(self, command_list, server):
        assert type(command_list) == list or type(command_list) == tuple, '\n* Input must be a list or tuple'
        assert server in self.kc_available_server_list, '\n* Server %s not in active server list' % server
        clnt = SimpleClient(server, self.kc_pysces_port, self.kc_block_size)
        clnt.send(command_list)
        self.kc_tentacle_response[server] = clnt.response
        ##  print self.kc_tentacle_response

    def sendJobToOne(self, command, server):
        assert server in self.kc_available_server_list, '\n* Server %s not in active server list' % server
        self.kc_tentacle_process_map[server] = ThreadedClient(command, server, self.kc_pysces_port, self.kc_block_size)

    def sendJobToAll(self, command):
        for server in self.kc_available_server_list:
            self.sendJobToOne(command, server)

    def runAllJobs(self):
        Istarted = []
        for s in list(self.kc_tentacle_process_map.keys()):
            if self.kc_tentacle_process_map[s] != None:
                if not self.kc_tentacle_process_map[s].isAlive():
                    self.kc_tentacle_process_map[s].start()
                    Istarted.append(self.kc_tentacle_process_map[s])
        for s in Istarted:
            s.join()
        del Istarted

    def clearProcessMap(self):
        for s in list(self.kc_tentacle_process_map.keys()):
            if self.kc_tentacle_process_map[s] != None:
                if not self.kc_tentacle_process_map[s].isAlive():
                    self.kc_tentacle_process_map[s] = None
        ##  print self.kc_tentacle_process_map


    def getDataFromOneJob(self, server):
        assert server in self.kc_tentacle_process_map, '\n* Server %s not in active server list' % server
        if self.kc_tentacle_process_map[server].response == 'True':
            print('* \"%s\" reports completing job without error' % server)
        else:
            print('* Check result! \"%s\" reports job processing failure'  % server)
        client = SimpleMultiReadClient(server, self.kc_pysces_port, self.kc_block_size)
        client.send('P_GETDATA')
        self.kc_tentacle_response[server] = cPickle.loads(client.response)
        print('\n', server, 'returns', type(self.kc_tentacle_response[server]))
        print('Data successfully retrieved from: %s \n' % server)

    def getDataFromAllJobs(self, pause=0.2):
        for s in list(self.kc_tentacle_process_map.keys()):
            if self.kc_tentacle_process_map[s] != None:
                self.getDataFromOneJob(s)
                time.sleep(pause)

    def killTentacles(self):
        GO = True
        while GO:
            inp = input('\nYou have initiated tentacle shutdown ... are you sure? (y/n): ')
            if inp in ['y','Y','yes']:
                GO = False
                for server in self.kc_available_server_list:
                    SimpleClient(server, self.kc_status_port, self.kc_block_size).send(['KILL'])
                    SimpleClient(server, self.kc_pysces_port, self.kc_block_size).send(['KILL'])
                print('Warning: all servers deleted for this host')
                self.kc_available_server_list = []
            elif inp in ['n','N','no']:
                GO = False
            else:
                print('\nPlease enter y for yes or n for no')

    def chkpsc(self,F):
        try:
            if F[-4:] == '.psc':
                pass
            else:
                print('Assuming extension is .psc')
                F += '.psc'
        except:
            print('Chkpsc error')
        return F

    def startModelServer(self, model_file, model_dir):
        self.model_file = model_file
        self.model_dir = model_dir
        self.model_file = self.chkpsc(self.model_file)
        mpath = os.path.join(self.model_dir, self.model_file)
        if os.path.exists(mpath):
            try:
                self.kc_model_server = ModelFileServer(CFSERVE_PORT, BLOCK_SIZE, 'model_server')
                self.kc_model_server.ReadFile(self.model_file, self.model_dir)
                self.kc_model_server.start()
            except Exception as ex:
                print('Server might already be running on this machine')
                print(ex)
        else:
            print('Error: file %s does not exist!\n' % mpath)
            raise NameError

    def startTentacleMonitor(self, interval=60):
        self.kc_tentacle_monitor = ServerStatusCheck(self.kc_available_server_list, self.kc_status_port, self.kc_block_size, interval)
        self.kc_tentacle_monitor.start()

    def ResetServersToReady(self):
        for server in self.kc_available_server_list:
            SimpleClient(server, self.kc_status_port, self.kc_block_size).send(['P_RESET_STATUS'])


class KrakenScanController(KrakenController):
    job_list = None
    JobsWaiting = None
    result_list = None
    result_array = None
    task_id = None
    working_dir = None
    Model_File = None
    Model_Dir = None
    tentacle_monitor_interval = 120
    init_commands = None
    once_off_tentacle_init = None

    def setEnv(self, task_id, working_dir, model_file, model_dir):
        self.task_id = task_id
        self.working_dir = working_dir
        self.job_list = []
        self.result_list = []
        self.Model_File = model_file
        self.Model_Dir = model_dir

    def startController(self):
        ##  self.blob = getBasicController()
        self.initialiseTentacles(directory_name=self.working_dir)
        self.getActiveTentacles()
        self.startModelServer(self.Model_File, self.Model_Dir)
        self.startTentacleMonitor(interval=self.tentacle_monitor_interval)
        print('Server list: %s\n' % self.kc_available_server_list)

    def setInitCmds(self, lst):
        assert type(lst) == list, 'This must be a list'
        self.init_commands = lst

    def setOnceOffTentacleInit(self, lst):
        assert type(lst) == list, 'This must be a list'
        self.once_off_tentacle_init = lst

    def setJobList(self, lst):
        assert type(lst) == list, 'This must be a list'
        self.job_list = lst


    def setScanJobs(self, start, end, intervals, job, log=False):
        """Splits a range into a number of jobs with intervals"""
        assert intervals >= 1, '\n* Minimum of 1 interval'
        if log:
            kpoints = scipy.logspace(scipy.log10(start), scipy.log10(end), intervals+1)
        else:
            kpoints = scipy.linspace(start, end, intervals+1)
        self.job_list = []
        for p in range(len(kpoints)-1):
            job2 = job % (kpoints[p], kpoints[p+1])
            self.job_list.append(job2)
            print(job2)

    def Run(self, raw_data_dump=True):
        START = time.time()
        self.Scheduler3(self.job_list, self.init_commands)
        MID = time.time()
        if raw_data_dump:
            try:
                self.Dump(self.result_list, self.task_id+'_raw_data')
            except Exception as ex:
                print('Raw data not saved', ex)
        print("Data generation time = %2.2f minutes." % ((MID-START)/60.0))

    def buildCycler(self, lst):
        """
        creates an infinite looping generator with itertools
        """
        return itertools.cycle(lst)

    def Dump(self, thing, fname):
        try:
            F = open(os.path.join(self.working_dir, fname.replace('.bin','')+'.bin'), 'wb')
            cPickle.dump(thing, F, 2)
            F.flush()
            F.close()
        except Exception as ex:
            print('Dump exception raised')
            print(ex)

    def deltaCommand(self):
        """
        This is used to setup conditions for each next-job-on-server
        """
        delta = 'P_NONE'
        return [delta]

    def Scheduler3(self, job_list, init_list):
        """
        job_list is a list of jobs to run
        init_list set of `static' command used to initialise each job
        """
        original_job_list = tuple(job_list)
        TIME_START = time.time()
        self.result_list = []
        self.JobsWaiting = True
        getServer = self.buildCycler(self.kc_available_server_list)
        if self.once_off_tentacle_init != None:
            self.sendCmdListToAll(self.once_off_tentacle_init, pause=0.5)
        while self.JobsWaiting:
            deferred_jobs = []
            job_servers = []
            for job in range(len(job_list)):
                if job < len(self.kc_available_server_list):
                    print('\nJob %s queued for processing...' % (job+1))
                    server = next(getServer)
                    job_servers.append(server)
                    print("\nProcessing job %s init list ..." % (job+1))
                    self.sendCmdListToOne(init_list, server)
                    print("\nProcessing job %s delta command ..." % (job+1))
                    self.sendCmdListToOne(self.deltaCommand(), server)
                    self.sendJobToOne(job_list[job], server)
                else:
                    deferred_jobs.append(job_list[job])
                    print('Job %s deferred and rescheduled.' % (job+1))
            print("\nProcessing queued jobs ...")
            self.runAllJobs()
            for server in job_servers:
                self.getDataFromOneJob(server)
                arr = self.kc_tentacle_response[server]
                self.result_list.append(arr)
                print(type(arr))
            self.clearProcessMap()
            if len(deferred_jobs) > 0:
                print('\nDeferred:', deferred_jobs)
                job_list = deferred_jobs
            else:
                self.JobsWaiting = False

        TIME_END = time.time()

        print('\nNumber of results = %s' % len(self.result_list))
        total_states = 0
        ##  for x in range(len(self.result_list)):
            ##  try:
                ##  print '\tResult %s has %s rows and %s columns' % (x+1, self.result_list[x].shape[0], self.result_list[x].shape[1])
                ##  total_states += self.result_list[x].shape[0]
            ##  except Exception, ex:
                ##  print type(self.result_list[x][0])
                ##  print '\tResult %s %s: %s' % (x+1, type(self.result_list[x][0]), self.result_list[x][0])
                ##  print original_job_list[x]
                ##  print ex
        ##  print 'Total time taken to complete %s state scan %s minutes' % (total_states, (TIME_END-TIME_START)/60.0)

    #investigate using vstack with type/range checking
    def concatenateResults(self):
        output = None
        for arr in range(len(self.result_list)):
            if type(self.result_list[arr]) == numpy.ndarray:
                if arr == 0:
                    output = self.result_list[arr]
                else:
                    output = scipy.concatenate((output,self.result_list[arr]))
        self.result_array = output

    def getResultArray(self):
        return self.result_array

    def getResultList(self):
        return self.result_list

    def getAndClearResultArray(self):
        print('ReSetting result_array and result_list (returning result_array)')
        tmp = self.result_array
        self.result_list = []
        self.result_array = None
        return tmp

    def getAndClearResultList(self):
        print('ReSetting result_array and result_list (returning result_list)')
        tmp = self.result_list
        self.result_list = []
        self.result_array = None
        return tmp



def getBasicController(lvl=3, directory=os.getcwd()):
    if lvl >= 1: blob = KrakenController()
    if lvl >= 2: blob.initialiseTentacles(directory_name=directory)
    if lvl >= 3: blob.getActiveTentacles()
    return blob




#!/usr/bin/env python
# Testing the new parallel scanner class - detection of invalid states
# --jr20101218

import os
backupdir = os.getcwd()

import numpy as np
import pysces

tbox=pysces.PyscesUtils.TimerBox()
import time

m=pysces.model('isola2a')

ser = pysces.Scanner(m)

print("Serial execution...")
print("Start: ", tbox.normal_timer('SER'))
print(next(tbox.SER))
t1=time.time()
ser.quietRun = True
ser.addScanParameter('V4',0.01,200,2000,log=True)
ser.addUserOutput('J_R1', 'A')
ser.Run()
print("Done: ", next(tbox.SER))
t2=time.time()
print("Duration: %.2f seconds" % (t2-t1))
ser.statespersecond = len(ser.ScanSpace)/(t2-t1)
print("States per second: %.1f" % ser.statespersecond)

print("\n\nParallel execution...scans per run =", 36)
par = pysces.ParScanner(m)
par.scans_per_run = 36
t3=time.time()
par.addScanParameter('V4',0.01,200,2000,log=True)
par.addUserOutput('J_R1', 'A')
par.Run()
t4=time.time()
print("Duration: %.2f seconds" % (t4-t3))
par.statespersecond = par.Tsteps/(t4-t3)
print("States per second: %.1f" % par.statespersecond)

print("\n Speedup: %.2f" % (par.statespersecond/ser.statespersecond))

print("\n\nParallel execution...with scatter and gather")
par2 = pysces.ParScanner(m, engine='ipcluster')
t5=time.time()
par2.addScanParameter('V4',0.01,200,2000,log=True)
par2.addUserOutput('J_R1', 'A')
par2.RunScatter()
t6=time.time()
print("Duration: %.2f seconds" % (t6-t5))
par2.statespersecond = par2.Tsteps/(t6-t5)
print("States per second: %.1f" % par2.statespersecond)

print("\n Speedup: %.2f" % (par2.statespersecond/ser.statespersecond))

print("\n===========\nComparing results...")
sss = ser.SteadyStateResults
pss = par.SteadyStateResults
suo = ser.UserOutputResults
puo = par.UserOutputResults
p2ss = par2.SteadyStateResults
p2uo = par2.UserOutputResults

print("\nTesting nan's:")
print("serial and parallel (tc):      ", \
np.alltrue(np.where(np.isnan(sss))[0] == np.where(np.isnan(pss))[0]), \
np.alltrue(np.where(np.isnan(sss))[1] == np.where(np.isnan(pss))[1]), \
np.alltrue(np.where(np.isnan(suo))[0] == np.where(np.isnan(puo))[0]), \
np.alltrue(np.where(np.isnan(suo))[1] == np.where(np.isnan(puo))[1]))
print("serial and parallel (scatter): ", \
np.alltrue(np.where(np.isnan(sss))[0] == np.where(np.isnan(p2ss))[0]), \
np.alltrue(np.where(np.isnan(sss))[1] == np.where(np.isnan(p2ss))[1]), \
np.alltrue(np.where(np.isnan(suo))[0] == np.where(np.isnan(p2uo))[0]), \
np.alltrue(np.where(np.isnan(suo))[1] == np.where(np.isnan(p2uo))[1]))

print("\nTesting finite values:")
print("serial and parallel (tc):      ", \
np.alltrue(sss[np.where(np.isfinite(sss))]==pss[np.where(np.isfinite(pss))]), \
np.alltrue(suo[np.where(np.isfinite(suo))]==puo[np.where(np.isfinite(puo))]))
print("serial and parallel (scatter): ", \
np.alltrue(sss[np.where(np.isfinite(sss))]==p2ss[np.where(np.isfinite(p2ss))]), \
np.alltrue(suo[np.where(np.isfinite(suo))]==p2uo[np.where(np.isfinite(p2uo))]))


os.chdir(backupdir)


"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = '''
            PyscesUtils
            -----------

            The PyscesUtils module holds a collection of methods of general use such as timers and
            array export fucntionality that can be accessed as **pysces.write.**

            '''

import fileinput
import os, sys
import time


def VersionCheck(ver='0.1.5'):
    print("dead")


def str2bool(s):
    """
    Tries to convert a string to a Python boolean

    - *s* True if 'True', 'true' or'1' else False

    """
    if s in ['True', 'true', '1']:
        return True
    else:
        return False


class TimerBox:
    """
    A timer "container" class that can be used to hold user defined timers

    """

    __go__ = 1

    def __init__(self):
        if os.sys.version_info[0] < 2 or os.sys.version_info[1] < 3:
            print(
                'Your version of Python ('
                + str(os.sys.version_info[:3])
                + ') might be too old\
            to use this class. Python 2.2 or newer is required'
            )

    def normal_timer(self, name):
        """
        normal_timer(name)

        Creates a normal timer method with <name> in the TimerBox instance. Normal timers
        print the elapsed time since creation when called.

         - *name* the timer name

        """
        assert type(name) == str or type(name) == unicode

        def timer(startTime=time.time()):
            while self.__go__:
                ms = divmod(time.time() - startTime, 60)
                nowTime = str(int(ms[0])) + ' min ' + str(int(ms[1])) + ' sec'
                yield nowTime

        setattr(self, name, timer())

    def step_timer(self, name, maxsteps):
        """
        step_timer(name,maxsteps)

        Creates a step timer method with <name> in the TimerBox instance. Step timers
        print the elapsed time as well as the next step out of maxsteps when called.

        - *name* the timer name
        - *maxsteps* the maximum number of steps associated with this timer

        """
        assert type(name) == str, 'Name is a string representing the timer name'
        assert type(maxsteps) == int or type(maxsteps) == float, 'int or float needed'

        setattr(self, name + '_cstep', 0)

        def timer(startTime=time.time(), maxS=maxsteps):
            while self.__go__:
                ms = divmod(time.time() - startTime, 60)
                nowStep = (
                    'step '
                    + str(getattr(self, name + '_cstep'))
                    + ' of '
                    + str(maxS)
                    + ' ('
                    + str(int(ms[0]))
                    + ' min '
                    + str(int(ms[1]))
                    + ' sec)'
                )
                yield nowStep

        setattr(self, name, timer())

    def stop(self, name):
        """
        Delete the timer <name> from the TimerBox instance

        - *name* the timer name to delete

        """
        delattr(self, name)
        try:
            getattr(self, name + '_cstep')
            delattr(self, name + '_cstep')
        except:
            pass

    def reset_step(self, name):
        """
        Reset the number of steps of timer <name> in the TimerBox to zero

        - *name* the step timer whose steps should be reset

        """
        try:
            getattr(self, name + '_cstep')
            setattr(self, name + '_cstep', 0)
        except:
            pass


def CopyTestModels(*args, **kwargs):
    print("moved to PyscesTest")


def CopyModels(*args, **kwargs):
    print("dead")


def ConvertFileD2U(Filelist):
    """
    Converts a [Filename] from rn to n inplace no effect if the line termination is correct

    - *Filelist* a file or list of files to convert

    """
    for line in fileinput.input(Filelist, inplace=1):
        try:
            if line[-2] == '\r':
                print(line[:-2] + '\n', end=' ')
            else:
                print(line, end=' ')
        except:
            print(line, end=' ')


def ConvertFileU2D(Filelist):
    """
    Converts a [Filename] from n to rn inplace no effect if the line termination is correct:

    - *Filelist* a file or list of files to convert

    """
    for line in fileinput.input(Filelist, inplace=1):
        try:
            if line[-2] != '\r':
                print(line[:-1] + '\n', end=' ')
            else:
                print(line, end=' ')
        except:
            print('\n', end=' ')


class WriteOutput(object):
    """
    This code is adapted from:

    CBMPy: CBTools module

    Constraint Based Modelling in Python (http://cbmpy.sourceforge.net)
    Copyright (C) 2009-2017 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

    """

    def exportLabelledArray(self, arr, names, fname, sep=',', sformat='%f'):
        """
        Write a 2D array type object to file

         - *arr* the an array like object
         - *names* the list of row names
         - *fname* the output filename
         - *sep* [default=','] the column separator
         - *format* [default='%s'] the output number format

        """
        if names != None:
            assert arr.shape[0] == len(
                names
            ), '\n ...  rows must equal number of names!'
        F = open(fname, 'w')
        cntr = 0
        for r in range(arr.shape[0]):
            if names != None:
                F.write(('{}' + sep).format(names[r]))
            for c in range(arr.shape[1]):
                if c < arr.shape[1] - 1:
                    F.write((sformat + sep) % arr[r, c])
                else:
                    F.write((sformat + '\n') % arr[r, c])
            cntr += 1
            if cntr >= 250:
                F.flush()
                cntr = 1
        F.write('\n')
        F.flush()
        F.close()
        print('exported to %s' % fname)

    def exportLabelledArrayWithHeader(
        self, arr, names, header, fname, sep=',', format='%f'
    ):
        """
        Export an array with row names and header

         - *arr* the an array like object
         - *names* the list of row names
         - *header* the list of column names
         - *fname* the output filename
         - *sep* [default=','] the column separator
         - *format* [default='%s'] the output number format
         - *appendlist* [default=False] if True append the array to *fname* otherwise create a new file

        """
        if names != None:
            assert arr.shape[0] == len(
                names
            ), '\n ...  rows must equal number of names!'
        if header != None:
            assert arr.shape[1] == len(
                header
            ), '\n ...  cols must equal number of header names!'
        F = open(fname, 'w')
        cntr = 0
        if header != None:
            if names != None:
                hstr = ' ' + sep
            else:
                hstr = ''
            for h in header:
                hstr += str(h) + sep
            hstr = hstr[:-1] + '\n'
            F.write(hstr)
            del hstr
        for r in range(arr.shape[0]):
            if names != None:
                F.write(('%s' + sep) % names[r])
            for c in range(arr.shape[1]):
                if c < arr.shape[1] - 1:
                    F.write((format + sep) % arr[r, c])
                else:
                    F.write((format + '\n') % arr[r, c])
            cntr += 1
            if cntr >= 250:
                F.flush()
                cntr = 1
        F.write('\n')
        F.flush()
        F.close()
        print('exported to %s' % fname)

    def exportLabelledLinkedList(
        self, arr, fname, names=None, sep=',', format='%s', appendlist=False
    ):
        """
        Write a 2D linked list [[...],[...],[...],[...]] and optionally a list of row labels to file:

         - *arr* the linked list
         - *fname* the output filename
         - *names* the list of row names
         - *sep* [default=','] the column separator
         - *format* [default='%s'] the output number format
         - *appendlist* [default=False] if True append the array to *fname* otherwise create a new file

        """
        if names != None:
            assert len(arr) == len(names), '\n ...  rows must equal number of names!'
        if not appendlist:
            F = open(fname, 'w')
        else:
            F = open(fname, 'a')
        cntr = 0
        for r in range(len(arr)):
            if names != None:
                F.write(('%s' + sep) % names[r])
            col_l = len(arr[0])
            for c in range(col_l):
                if c < col_l - 1:
                    if arr[r][c] == 0.0:
                        F.write('0.0' + sep)
                    else:
                        try:
                            F.write((format + sep) % arr[r][c])
                        except UnicodeEncodeError:
                            F.write((format + sep) % 'uError')
                else:
                    if arr[r][c] == 0.0:
                        F.write('0.0\n')
                    else:
                        try:
                            F.write((format + '\n') % arr[r][c])
                        except UnicodeEncodeError:
                            F.write((format + '\n') % 'uError')
            cntr += 1
            if cntr >= 250:
                F.flush()
                cntr = 1
        ##  F.write('\n')
        F.flush()
        F.close()
        del arr
        if not appendlist:
            print('exported to %s' % fname)

    def exportLabelledArrayWithHeader2CSV(self, arr, names, header, fname):
        """
        Export an array with row names and header to fname.csv

         - *arr* the an array like object
         - *names* the list of row names
         - *header* the list of column names
         - *fname* the output filename

        """
        fname += '.csv'
        self.exportLabelledArrayWithHeader(
            arr, names, header, fname, sep=',', format='%f'
        )

    def exportLabelledArray2CSV(self, arr, names, fname):
        """
        Export an array with row names to fname.csv

         - *arr* the an array like object
         - *names* the list of row names
         - *fname* the output filename

        """
        fname += '.csv'
        self.exportLabelledArray(arr, names, fname, sep=',', sformat='%f')

    def exportArray2CSV(self, arr, fname):
        """
        Export an array to fname.csv

         - *arr* the an array like object
         - *fname* the output filename
         - *sep* [default=','] the column separator

        """
        fname += '.csv'
        self.exportLabelledArray(arr, None, fname, sep=',', sformat='%f')

    def exportLabelledArrayWithHeader2TXT(self, arr, names, header, fname):
        """
        Export an array with row names and header to fname.txt

         - *arr* the an array like object
         - *names* the list of row names
         - *header* the list of column names
         - *fname* the output filename

        """
        fname += '.txt'
        self.exportLabelledArrayWithHeader(
            arr, names, header, fname, sep='\t', format='%f'
        )

    def exportLabelledArray2TXT(self, arr, names, fname):
        """
        Export an array with row names to fname.txt

         - *arr* the an array like object
         - *names* the list of row names
         - *fname* the output filename

        """
        fname += '.txt'
        self.exportLabelledArray(arr, names, fname, sep='\t', sformat='%f')

    def exportArray2TXT(self, arr, fname):
        """
        Export an array to fname.txt

         - *arr* the an array like object
         - *fname* the output filename
         - *sep* [default=','] the column separator

        """
        fname += '.txt'
        self.exportLabelledArray(arr, None, fname, sep='\t', sformat='%f')



import os
import scipy

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

from Kraken import time, KrakenController, getBasicController, BLOCK_SIZE

blob = getBasicController()

kin = input('\nKill all idle tentacles? (y/n):\n')
if kin in ['y', 'yes']:
    blob.killTentacles()



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__

__doc__ = "PySCeS parser module -- uses  PLY 1.5 or newer"


import os, copy
from .lib import lex
from .lib import yacc
from getpass import getuser
from time import sleep, strftime

try:
    from importlib import reload  # Python 3
except ImportError:
    pass

# use net stoichiometry, disable for StomPy
__USE_NET_STOICH__ = True

RESERVEDTERMS = (
    'Vvec',
    '__CDvarUpString__',
    '__SALL__',
    '__SI__',
    '__epmatrix__',
    '__evmatrix__',
    '__fixed_species__',
    '__kmatrix__',
    '__kzeromatrix__',
    '__lmatrix__',
    '__lzeromatrix__',
    '__mapFunc_R__',
    '__mapFunc__',
    '__mode_suppress_info__',
    '__modifiers__',
    '__nmatrix__',
    '__nrmatrix__',
    '__parameters__',
    '__reactions__',
    '__s_varFunc__',
    '__species__',
    '__vFunc__',
    'cc_all',
    'cc_all_col',
    'cc_all_row',
    'cc_conc',
    'cc_conc_col',
    'cc_conc_row',
    'cc_flux',
    'cc_flux_col',
    'cc_flux_row',
    'conservation_matrix',
    'conservation_matrix_col',
    'conservation_matrix_row',
    'display_debug',
    'eModeExe_dbl',
    'eModeExe_int',
    'elas_epar_upsymb',
    'elas_evar_upsymb',
    'elas_par',
    'elas_par_col',
    'elas_par_row',
    'elas_par_u',
    'elas_var',
    'elas_var_col',
    'elas_var_row',
    'elas_var_u',
    'emode_file',
    'emode_intmode',
    'emode_userout',
    'fintslv_range',
    'fintslv_rmult',
    'fintslv_step',
    'fintslv_tol',
    'fixed_species',
    'hybrd_epsfcn',
    'hybrd_factor',
    'hybrd_maxfev',
    'hybrd_mesg',
    'hybrd_xtol',
    'info_flux_conserve',
    'info_moiety_conserve',
    'init_species_array',
    'kel_unscaled',
    'kmatrix',
    'kmatrix_col',
    'kmatrix_row',
    'kmatrix_scaled',
    'kzeromatrix',
    'kzeromatrix_col',
    'kzeromatrix_row',
    'lmatrix',
    'lmatrix_col',
    'lmatrix_row',
    'lmatrix_scaled',
    'lsoda_atol',
    'lsoda_h0',
    'lsoda_hmax',
    'lsoda_hmin',
    'lsoda_mesg',
    'lsoda_mxordn',
    'lsoda_mxords',
    'lsoda_mxstep',
    'lsoda_rtol',
    'lzeromatrix',
    'lzeromatrix_col',
    'lzeromatrix_row',
    'mach_floateps',
    'mca_ccall_altout',
    'mca_ccall_concout',
    'mca_ccall_fluxout',
    'mca_ccj_upsymb',
    'mca_ccs_upsymb',
    'mca_ci',
    'mca_ci_col',
    'mca_ci_row',
    'mca_cjd',
    'mca_cjd_col',
    'mca_cjd_row',
    'mca_csd',
    'mca_csd_col',
    'mca_csd_row',
    'misc_write_arr_lflush',
    'mode_eigen_output',
    'mode_elas_deriv',
    'mode_elas_deriv_dx',
    'mode_elas_deriv_factor',
    'mode_elas_deriv_min',
    'mode_elas_deriv_order',
    'mode_mca_scaled',
    'mode_number_format',
    'mode_sim_init',
    'mode_sim_max_iter',
    'mode_solver',
    'mode_solver_fallback',
    'mode_solver_fallback_integration',
    'mode_state_init',
    'mode_state_init2_array',
    'mode_state_init3_factor',
    'mode_state_mesg',
    'modifiers',
    'nleq2_ibdamp',
    'nleq2_iormon',
    'nleq2_iscal',
    'nleq2_iter',
    'nleq2_jacgen',
    'nleq2_mesg',
    'nleq2_mprerr',
    'nleq2_nonlin',
    'nleq2_qnscal',
    'nleq2_qrank1',
    'nleq2_rtol',
    'nmatrix',
    'nmatrix_col',
    'nmatrix_row',
    'nrmatrix',
    'nrmatrix_col',
    'nrmatrix_row',
    'parameters',
    'pitcon_abs_tol',
    'pitcon_allow_badstate',
    'pitcon_filter_neg',
    'pitcon_filter_neg_res',
    'pitcon_fix_small',
    'pitcon_flux_gen',
    'pitcon_init_par',
    'pitcon_iter',
    'pitcon_jac_opt',
    'pitcon_jac_upd',
    'pitcon_limit_point_idx',
    'pitcon_limit_points',
    'pitcon_max_grow',
    'pitcon_max_step',
    'pitcon_max_steps',
    'pitcon_min_step',
    'pitcon_output_lvl',
    'pitcon_par_opt',
    'pitcon_par_space',
    'pitcon_rel_tol',
    'pitcon_start_dir',
    'pitcon_start_step',
    'pitcon_targ_val',
    'pitcon_targ_val_idx',
    'pitcon_target_points',
    'reactions',
    'scan1_dropbad',
    'scan1_mca_mode',
    'scan_in',
    'scan_out',
    'scan_res',
    'sim_end',
    'sim_points',
    'sim_res',
    'sim_start',
    'sim_time',
    'species',
    'state_flux',
    'state_set_conserve',
    'state_species',
    'write_array_header',
    'write_array_html_footer',
    'write_array_html_format',
    'write_array_html_header',
    'write_array_spacer',
    'zero_val',
)


class PySCeSParser:
    """The original PySCeS parser has been rewritten and extended to include
    many features that allow for SBML compatibility such as MathML 2.0 support,
    function definitions, assignment/rate rules and events."""

    MathmlToNumpy_funcs = {
        'pow': 'pow',
        'root': 'pow',
        'abs': 'abs',
        'exp': 'math.exp',
        'ln': 'math.log',
        'log': 'math.log10',
        'floor': 'numpy.floor',
        'ceiling': 'numpy.ceil',
        'factorial': None,
        'sin': 'numpy.sin',
        'cos': 'numpy.cos',
        'tan': 'numpy.tan',
        'sec': None,
        'csc': None,
        'cot': None,
        'sinh': 'numpy.sinh',
        'cosh': 'numpy.cosh',
        'tanh': 'numpy.tanh',
        'sech': None,
        'csch': None,
        'coth': None,
        'arcsin': 'numpy.arcsin',
        'arccos': 'numpy.arccos',
        'arctan': 'numpy.arctan',
        'arcsec': None,
        'arccsc': None,
        'arccot': None,
        'arcsinh': 'numpy.arcsinh',
        'arccosh': 'numpy.arccosh',
        'arctanh': 'numpy.arctanh',
        'arcsech': None,
        'arccsch': None,
        'arccoth': None,
        'eq': 'operator.eq',
        'neq': 'operator.ne',
        'gt': 'operator.gt',
        'geq': 'operator.ge',
        'lt': 'operator.lt',
        'leq': 'operator.le',
        'ceil': 'numpy.ceil',
        'sqrt': 'numpy.sqrt',  # libsbml aliases
        'equal': 'operator.eq',
        'not_equal': 'operator.ne',  # numpy2numpy aliases
        'greater': 'operator.gt',
        'greater_equal': 'operator.ge',  # numpy2numpy aliases
        'less': 'operator.lt',
        'less_equal': 'operator.le',  # numpy2numpy aliases
        'ne': 'operator.ne',
        'ge': 'operator.ge',
        'le': 'operator.le',  # operator2operator
        'xor': 'operator.xor',
        'piecewise': 'self._piecewise_',
        '_piecewise_': 'self._piecewise_',
        'not': 'operator.not_',
        'not_': 'operator.not_',
        'max': 'max',
        'min': 'min',
    }

    MathmlToNumpy_symb = {
        'notanumber': 'numpy.NaN',
        'pi': 'numpy.pi',
        'infinity': 'numpy.Infinity',
        'exponentiale': 'numpy.e',
        'true': 'True',
        'false': 'False',
        'True': 'True',
        'False': 'False',
    }

    SymbolReplacements = None

    MathmlToInfix = {
        'and': 'and',
        'or': 'or',
        'true': 'True',
        'false': 'False',
        'xor': 'xor',
    }

    BOOLEANTRUE = ('True', 'TRUE', 'true')
    BOOLEANFALSE = ('False', 'False', 'false')

    precedence = (
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIVIDE'),
        ('left', 'POWER'),
        ('right', 'UMINUS'),
    )

    # List of token names
    tokens = (
        'FIXDEC',
        'FUNCDEC',
        'EVENTDEC',
        'COMPDEC',
        'RRULEDEC',
        'OBJFUNCDEC',
        'FLUXBNDDEC',
        'USERCNSTRDEC',
        'IRREV',
        'REAL',
        'INT',
        'PLUS',
        'MINUS',
        'TIMES',
        'DIVIDE',
        'POWER',
        'LPAREN',
        'RPAREN',
        'EQUALS',
        'SYMBEQUI',
        'COMMA',
        'REACTION_ID',
        'STOICH_COEF',
        'NAME',
        'FORCEFUNC',
        'TIMEFUNC',
        'USERFUNC',
        'INITFUNC',
        'IN',
        'POOL',
        'KEYWORD',
        'KEYWORDB',
        'UNIT',
        'MULTICOMMENT',
    )

    ReservedTerms = RESERVEDTERMS

    #################################################################
    #################################################################

    def __init__(self, debug=0):
        self.display_debug = debug
        # Lists
        self.LexErrors = []  # List of lexing errors
        self.ParseErrors = []
        self.SymbolErrors = []
        self.ParseOK = True
        self.LexOK = True
        self.ModelUsesNumpyFuncs = False

        self.ReactionIDs = []  # List of reaction names
        self.Names = []  # List of all reagent, parameter and function names
        self.nDict = {}  # Dictionary containing all reaction information
        self.InDict = {}  # spatial dictionary
        self.sDict = {}
        self.pDict = {}  # parameter dict
        self.uDict = {
            'substance': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'mole'},
            'volume': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'litre'},
            'time': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'second'},
            'length': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'metre'},
            'area': {'exponent': 2, 'multiplier': 1.0, 'scale': 0, 'kind': 'metre'},
        }
        self.KeyWords = {
            'Modelname': None,
            'Description': None,
            'Species_In_Conc': None,
            'Output_In_Conc': None,
        }
        self.compartments = {}
        self.InitStrings = []  # Initialisation strings
        self.Inits = []  # Initialised entities
        self.Reagents = []  # All reagents found during parsing of reactions
        self.species = []  # Variable reagents that occur in reactions
        self.FixedReagents = []  # Fixed reagents
        self.ReacParams = []  # Temporary list of reaction parameters
        self.InitParams = []  # Initialised parameters

        self.ForceFunc = []  # Forcing function definition
        self.TimeFunc = []  # Time function definition
        self.UserFunc = []  # User function definition
        self.InitFunc = []  # Intitialization function definition
        self.Functions = {}  # new Function blocks
        self.Events = {}  # new event blocks
        self.AssignmentRules = {}  # new forcing function blocks
        self.UserFuncs = {}  # new user function blocks
        self.ModelInit = {}  # new model initialisation blocks
        self.cbm_FluxBounds = {}  # flux bounds in a CB model
        self.cbm_ObjectiveFunctions = {}  # objective functions for a CB model
        self.cbm_UserFluxConstraints = {}  # objective functions for a CB model

        self.mach_spec_eps2k = 2.2204460492503131e-14
        self.AllRateEqsGiven = (
            1  # Flag to check that all rate equations have been given
        )
        self.Debug = 0

        # elementary regular expressions used as building blocks
        self.Int = r'\d+'  # Integer
        self.Dec = self.Int + '\.' + self.Int  # Decimal
        self.Exp = r'([E|e][\+|\-]?)' + self.Int  # Exponent
        self.Real = (
            self.Dec + '(' + self.Exp + ')?' + '|' + self.Int + self.Exp
        )  # Real - dec w/o optional exp or int with exp

        # Simple tokens
        self.t_IRREV = r'>'
        self.t_REAL = self.Real
        self.t_INT = self.Int
        self.t_PLUS = r'\+'
        self.t_MINUS = r'-'
        self.t_TIMES = r'\*'
        self.t_DIVIDE = r'/'
        self.t_POWER = '\*\*'
        self.t_LPAREN = r'\('
        self.t_RPAREN = r'\)'
        self.t_EQUALS = r'='
        self.t_COMMA = r','
        self.t_POOL = r'\$pool'
        self.t_IN = r'@'

    t_ignore = ' \t\r'  # Ignore spaces and tabs --- and windows return - brett 20040229

    def t_comment(self, t):
        r'\#.+\n'  # Match from # to newline
        t.lineno += 1  # Increment line number

    def t_multilinecomment(self, t):
        r'"""(.|\n)*?"""'
        t.type = 'MULTICOMMENT'

    def t_newline(self, t):
        r'\n+'  # Match newline
        t.lineno += len(t.value)  # Increment with number of consecutive newlines

    def t_SYMBEQUI(self, t):
        r'!=|<'
        t.type = 'SYMBEQUI'
        print('t', t)
        return t

    def t_FIXDEC(self, t):
        r'FIX:'
        t.type = 'FIXDEC'
        t.value = 'FIX:'
        return t

    def t_KW_Name(self, t):
        r'Modelname:.+\n'
        t.type = 'KEYWORD'
        return t

    def t_KW_Description(self, t):
        r'Description:(.| )+\n'
        t.type = 'KEYWORD'
        return t

    def t_KW_ModelType(self, t):
        r'ModelType:(.| )+\n'
        t.type = 'KEYWORD'
        return t

    def t_KW_Species_In_Conc(self, t):
        r'Species_In_Conc:.+\n'
        t.type = 'KEYWORDB'
        return t

    def t_KW_Output_In_Conc(self, t):
        r'Output_In_Conc:.+\n'
        t.type = 'KEYWORDB'
        return t

    def t_Unit(self, t):
        r'Unit(Substance|Time|Length|Area|Volume):.+\n'
        t.type = 'UNIT'
        return t

    def t_FUNCDEC(self, t):
        r'Function:(.|\n)*?}'
        t.type = 'FUNCDEC'
        return t

    def t_RATERULEDEC(self, t):
        r'RateRule:(.|\n)*?}'
        t.type = 'RRULEDEC'
        return t

    def t_EVENTDEC(self, t):
        r'Event:(.|\n)*?}'
        t.type = 'EVENTDEC'
        return t

    def t_OBJFUNCDEC(self, t):
        r'ObjectiveFunctions:(.|\n)*?}'
        t.type = 'OBJFUNCDEC'
        return t

    def t_FLUXBNDDEC(self, t):
        r'FluxBounds:(.|\n)*?}'
        t.type = 'FLUXBNDDEC'
        return t

    def t_USERCNSTRDEC(self, t):
        r'UserFluxConstraints:(.|\n)*?}'
        t.type = 'USERCNSTRDEC'
        return t

    def t_COMPDEC(self, t):
        r'Compartment:.+\n'  # match from cc<space> to end of line
        t.type = 'COMPDEC'
        return t

    def t_FORCEFUNC(self, t):
        r'!F\ .+\n'  # match from !F<space> to end of line
        t.type = 'FORCEFUNC'
        t.lineno += 1  # Increment line number
        t.value = t.value[3:]
        return t

    def t_TIMEFUNC(self, t):
        r'!T\ .+\n'  # match from !T<space> to end of line
        t.type = 'TIMEFUNC'
        t.lineno += 1  # Increment line number
        t.value = t.value[3:]
        return t

    def t_USERFUNC(self, t):
        r'!U\ .+\n'  # match from !U<space> to end of line
        t.type = 'USERFUNC'
        t.lineno += 1  # Increment line number
        t.value = t.value[3:]
        return t

    def t_INITFUNC(self, t):
        r'!I\ .+\n'  # match from !I<space> to end of line
        t.type = 'INITFUNC'
        t.lineno += 1  # Increment line number
        t.value = t.value[3:]
        return t

    def t_REACTION_ID(self, t):
        r'[a-zA-Z_][\w@]*:'  # Match any letter in [a-zA-Z0-9_] up to a colon
        # allow "_" to startswith - brett 20050823
        t.type = 'REACTION_ID'
        t.value = t.value[:-1]  # remove the colon

        # print t.value, self.ReactionIDs
        # sleep(1)
        if '@' in t.value:
            ts = t.value.split('@')
            t.value = ts[0]
            self.InDict.update({ts[0]: ts[1]})
        if t.value in self.ReactionIDs:
            # brett I think this is hyperactive reporting ... removing 20050321
            # self.LexErrors.append(('Duplicate ReactionID ', t.lineno, t.value, t.type))
            pass
        else:
            self.ReactionIDs.append(t.value)
        # possible alternative to above - brett
        # if t.value not in self.ReactionIDs:
        #    self.ReactionIDs.append(t.value)
        return t

    def t_STOICH_COEF(self, t):
        r'\{[\d+|\.|E|e|\+|\-]+\}'

        # old regex upgraded to handle anything that resembles a number, might cause downstream problems
        # r'\{\d+\}|\{\d+\.\d+\}'
        t.type = 'STOICH_COEF'
        t.value = t.value[1:-1]  # Remove left and right curly brackets
        return t

    def t_NAME(self, t):
        r'numpy\.[\w]*|math\.[\w]*|operator\.[\w]*|random\.[\w]*|[a-zA-Z_][\w@]*'
        SciCons = False
        if '@' in t.value:
            ts = t.value.split('@')
            t.value = ts[0]
            self.InDict.update({ts[0]: ts[1]})
        if t.value in self.MathmlToNumpy_symb:
            if self.MathmlToNumpy_symb[t.value] == None:
                self.SymbolErrors.append(t.value)
                print('\nSymbol \"{}\" not yet supported by PySCeS.'.format(t.value))
                gt = 'unknown_symbol_' + t.value
                t.value = gt
            else:
                gt = self.MathmlToNumpy_symb[t.value]
                t.value = gt
            self.ModelUsesNumpyFuncs = 1
            SciCons = True
        elif self.SymbolReplacements != None and t.value in self.SymbolReplacements:
            if t.value not in self.Names:
                self.Names.append('self.' + self.SymbolReplacements[t.value])
            gt = 'self.' + self.SymbolReplacements[t.value]
            t.value = gt
        elif (
            t.value not in self.Names and t.value != '_TIME_'
        ):  # Only add to list if absent in list
            self.Names.append('self.' + t.value)
        if (
            t.value[:6] == 'numpy.'
            or t.value[:5] == 'math.'
            or t.value[:9] == 'operator.'
            or t.value[:7] == 'random.'
        ):
            pass
        elif (
            t.value not in self.MathmlToNumpy_funcs and not SciCons
        ):  # make class attributes, ignore function names
            gt = 'self.' + t.value
            t.value = gt
        t.type = 'NAME'
        return t

    def t_error(self, t):
        '''
        try:
            self.LexErrors.append(('Lexer error ', t.lineno, t.value, t.type))
        except:
            print 'Lexer error'
           #print 'Illegal character, Line ' + str(t.lineno) + ' :' + str(t.value[0])
        t.skip(1)
        '''
        print("Illegal character '{}'".format(t.value[0]))
        self.LexErrors.append(t.value[0])
        self.LexOK = False
        t.lexer.skip(1)

    ##############
    # The parser #
    ##############

    def Show(self, name, tok):
        if self.Debug:
            print(name, tok)

    def p_error(self, t):
        try:
            ##  self.ParseErrors.append(('Syntax error ', t.lineno, t.value, t.type))
            self.ParseErrors.append(t)
        except:
            print('p_error generated a parsing error')
        # fixes userwarning
        #tok = yacc.token()
        tok = self.__parser.token()
        while tok and tok.type != 'REACTION_ID':
            # fixes userwarning
            #tok = yacc.token()
            tok = self.__parser.token()
        self.ParseOK = False
        return tok

    def p_model(self, t):
        '''Model : Statement
                 | Model Statement '''

        self.Show('Model', t[0])

    def p_statement(self, t):
        '''Statement : Fixed
                     | FunctionDec
                     | RateRuleDec
                     | EventDec
                     | ObjFuncDec
                     | FluxBndDec
                     | UserCnstrDec
                     | CompartmentDec
                     | ReactionLine
                     | Initialise
                     | Forcedfunc
                     | Timefunc
                     | Userfunc
                     | Initfunc
                     | NameInName
                     | KeyWord
                     | KeyWordB
                     | Unit
                     | MultiComment
                     | SymbEqui'''
        self.Show('Statement', t[0])

    def p_nameinname(self, t):
        '''NameInName : NAME IN NAME'''
        print('This should never fire, what we need to do is ...')

    def p_inequalities_symb(self, t):
        '''SymbEqui : SYMBEQUI'''
        print('p_inequalities_symb', t[:])
        t[0] = t[1]

    def p_multicomment(self, t):
        '''MultiComment : MULTICOMMENT'''
        self.Show('KeyWord:', t[0])

    def p_keyword(self, t):
        '''KeyWord : KEYWORD'''
        ##  print 'KeyWord:', t[:]
        kpair = [e.strip() for e in t[1].split(':')]
        if kpair[0] in self.KeyWords:
            self.KeyWords.update({kpair[0]: kpair[1]})
        ##  print self.KeyWords
        self.Show('KeyWord:', t[0])

    def p_keywordboolean(self, t):
        '''KeyWordB : KEYWORDB'''
        ##  print 'KeyWordB:', t[:]
        kpair = [e.strip() for e in t[1].split(':')]
        if kpair[0] in self.KeyWords:
            if kpair[1] in self.BOOLEANTRUE:
                self.KeyWords.update({kpair[0]: True})
            elif kpair[1] in self.BOOLEANFALSE:
                self.KeyWords.update({kpair[0]: False})
        ##  print self.KeyWords
        self.Show('KeyWordB:', t[0])

    def p_unit(self, t):
        '''Unit : UNIT'''
        ##  print 'u', t[:]
        u = t[1].split(',')
        u[0] = u[0].split(':')
        u.append(u[0][1])
        u[0] = u[0][0].lower()
        ##  print u
        u[0] = u[0].replace('unit', '')
        ##  print u
        for i in range(len(u)):
            u[i] = u[i].strip()
            if i in [1, 2, 3]:
                u[i] = float(u[i])
        ##  print u
        self.uDict.update(
            {u[0]: {'multiplier': u[1], 'scale': u[2], 'exponent': u[3], 'kind': u[4]}}
        )
        self.Show('Unit:', t[0])

    def p_fixed(self, t):
        '''Fixed : FIXDEC FixList'''
        self.Show('Fixed:', t[0])

    def p_functiondec(self, t):
        '''FunctionDec : FUNCDEC'''
        ##  print 'p_functiondec', t[:]
        rawf = t[1].replace('Function:', '').lstrip()
        args = rawf[: rawf.find('{')].strip().split(',')
        name = args.pop(0)
        func = rawf[rawf.find('{') + 1 : rawf.find('}')].strip()
        args = [a.strip() for a in args]
        self.Functions.update({name: {'name': name, 'args': args, 'formula': func}})
        self.Show('FunctionDec:', t[0])

    def p_rateruledec(self, t):
        '''RateRuleDec : RRULEDEC'''
        ##  print 'p_functiondec', t[:]
        rawf = t[1].replace('RateRule:', '').lstrip()
        name = rawf[: rawf.find('{')].strip()
        func = rawf[rawf.find('{') + 1 : rawf.find('}')]
        self.AssignmentRules.update(
            {name: {'name': name, 'formula': func.strip(), 'type': 'rate'}}
        )
        self.Show('RateRuleDec:', t[0])

    def p_eventdec(self, t):
        '''EventDec : EVENTDEC'''
        rawf = t[1].replace('Event:', '').lstrip()
        args = rawf[: rawf.find('{')].strip().split(',')
        name = args.pop(0)
        delay = float(args.pop(-1))
        trigger = ''
        for a in args:
            trigger = trigger + a + ','
        trigger = trigger[:-1]

        rawF = rawf[rawf.find('{') + 1 : rawf.find('}')].split('\n')
        assignments = {}
        for ass in rawF:
            if len(ass.strip()) > 0:
                ass = ass.split('=')
                assignments.update({ass[0].strip(): ass[1].strip()})
        self.Events.update(
            {
                name: {
                    'delay': delay,
                    'name': name,
                    'trigger': trigger,
                    'assignments': assignments,
                    'tsymb': None,
                }
            }
        )
        self.Show('EventDec:', t[0])

    def p_objfuncdec(self, t):
        '''ObjFuncDec : OBJFUNCDEC'''
        rawf = t[1].replace('ObjectiveFunctions:', '').lstrip()
        args = rawf[: rawf.find('{')].strip().split(',')
        rawf = [
            r.strip() for r in rawf[rawf.find('{') + 1 : rawf.find('}')].split('\n')
        ]
        objectives = {}
        for oo in rawf:
            print(rawf)
            print(oo)
            if len(oo.strip()) > 0:
                oS = [t.strip() for t in oo.split(':')]
                active = False
                if oS[0] == args[0]:
                    active = True
                objectives.update(
                    {
                        oS[0]: {
                            'id': oS[0],
                            'active': active,
                            'osense': oS[1],
                            'fluxes': oS[2],
                        }
                    }
                )
        self.cbm_ObjectiveFunctions.update(objectives)
        self.Show('ObjFuncDec:', t[0])

    def p_fluxbnddec(self, t):
        '''FluxBndDec : FLUXBNDDEC'''
        rawf = t[1].replace('FluxBounds:', '').lstrip()
        rawf = [
            r.strip() for r in rawf[rawf.find('{') + 1 : rawf.find('}')].split('\n')
        ]

        fluxbnds = {}
        cntr = 0
        for oo in rawf:
            if len(oo.strip()) > 0:
                operator = None
                lhs = None
                rhs = None
                fid = 'fbnd_{}'.format(cntr)
                for oper in ['>=', '<=', '>', '<', '=']:
                    if oper in oo:
                        bnd = oo.split(oper)
                        operator = oper
                        lhs = bnd[0].strip()
                        rhs = float(bnd[1].strip())
                        break
                fluxbnds.update(
                    {fid: {'id': fid, 'flux': lhs, 'rhs': rhs, 'operator': operator}}
                )
                cntr += 1
        self.cbm_FluxBounds.update(fluxbnds)
        self.Show('FluxBndDec:', t[0])

    def p_usercnstrdec(self, t):
        '''UserCnstrDec : USERCNSTRDEC'''
        rawf = t[1].replace('UserFluxConstraints:', '').lstrip()
        rawf = [
            r.strip() for r in rawf[rawf.find('{') + 1 : rawf.find('}')].split('\n')
        ]

        fluxcnstr = {}
        for oo in rawf:
            if len(oo.strip()) > 0:
                oS = [t.strip() for t in oo.split(':')]
                id = oS[0]
                operator = None
                lhs = None
                rhs = None
                for oper in ['>=', '<=', '>', '<', '=']:
                    if oper in oS[1]:
                        cnstr = oS[1].split(oper)
                        operator = oper
                        lhs = [c.strip() for c in cnstr[0].split(',')]
                        lhs2 = []
                        for c2 in lhs:
                            c = c2.split(' ')
                            if len(c) == 1:
                                coeff = '+1'
                                val = c[0]
                            else:
                                coeff = c[0]
                                val = c[1]
                            lhs2.append((float(coeff), val))
                        rhs = float(cnstr[1].strip())
                        break
                fluxcnstr.update(
                    {
                        oS[0]: {
                            'id': oS[0],
                            'constraint': tuple(lhs2),
                            'rhs': rhs,
                            'operator': operator,
                        }
                    }
                )
        self.cbm_UserFluxConstraints.update(fluxcnstr)
        self.Show('UserCnstrDec:', t[0])

        self.Show('UserCnstrDec:', t[0])

    def p_compartmentdec(self, t):
        '''CompartmentDec : COMPDEC'''
        rawf = t[1].strip().replace('Compartment:', '')
        strAr = [st.strip() for st in rawf.split(',')]
        if len(strAr) <= 3:
            name = strAr[0]
            size = strAr[1]
            dimensions = strAr[2]
        if len(strAr) == 4:
            area = strAr[3]
        else:
            area = None

        self.compartments.update(
            {
                name: {
                    'name': name,
                    'size': float(size),
                    'dimensions': int(dimensions),
                    'compartment': None,
                    'area': None
                    ##  'volume' : None
                }
            }
        )

    def p_forcedfunc(self, t):
        '''Forcedfunc : FORCEFUNC'''
        self.ForceFunc.append(t[1])
        ar = t[1].split('=')
        name = ar[0].replace('self.', '').strip()
        func = ar[1].replace('self.', '').strip()
        self.AssignmentRules.update(
            {name: {'name': name, 'formula': func, 'type': 'assignment'}}
        )
        self.Show('Forcedfunc:', t[0])

    def p_timefunc(self, t):
        '''Timefunc : TIMEFUNC'''
        self.TimeFunc.append(t[1])
        self.Show('Timefunc:', t[0])

    def p_userfunc(self, t):
        '''Userfunc : USERFUNC'''
        self.UserFunc.append(t[1])
        ar = t[1].split('=')
        name = ar[0].replace('self.', '').strip()
        func = ar[1].replace('self.', '').strip()
        self.UserFuncs.update(
            {name: {'name': name, 'formula': func, 'type': 'userfuncs'}}
        )
        self.Show('Userfunc:', t[0])

    def p_initfunc(self, t):
        '''Initfunc : INITFUNC'''
        self.InitFunc.append(t[1])
        ar = t[1].split('=')
        name = ar[0].replace('self.', '').strip()
        value = ar[1].replace('self.', '').replace('\'', '').replace('"', '').strip()
        self.ModelInit.update({name: value})
        self.Show('Initfunc:', t[0])

    def p_fixedreagents(self, t):
        '''FixList : NAME
                   | NAME FixList'''
        if t[1] != None:
            self.FixedReagents.append(t[1])
        t[0] = [t[1]]
        try:
            t[0] += t[2]
        except:
            pass
        self.Show('FixList', t[0])

    # TODO:
    def p_initialise(self, t):
        '''Initialise : NAME EQUALS Expression'''
        value = None
        name = t[1].replace('self.', '')
        try:
            value = eval(t[3])
            # 20090402 we need to keep track of species parameter initialisations and then
            # perform a lookup or implement a proxy class that can store
            # evaluated assignments
        except Exception as ex:
            print('Initialisation error: {}'.format(t[3]))
            print(ex)

        self.sDict.update({name: {'name': name, 'initial': value}})
        t[0] = t[1] + t[2] + t[3]

        self.InitStrings.append(t[0])
        self.Inits.append(t[1])
        self.Show('Initialisation', t[0])

    def p_reaction_line(self, t):
        '''ReactionLine : REACTION_ID ReactionEq
                        | REACTION_ID ReactionEq Expression'''

        # global AllRateEqsGiven, self.ReacParams
        ReacID = t[1]
        if ReacID in self.nDict:
            self.ParseErrors.append(('Duplicate Reaction ', t.lineno, ReacID, None))
        self.nDict[ReacID] = {}  # Reaction dictionary for ReacID
        ##  self.nDict[ReacID]['Reagents'] = {} # Reagent dictionary within ReacID
        self.nDict[ReacID].update({'Reagents': {}, 'name': ReacID})
        # a list of all reagents specified in the stoichiometry, unmodified and not taken into consideration for anything
        self.nDict[ReacID].update({'AllReagents': t[2][0]})

        # brett: if an index exists sum the coefficients instead of adding a new one
        # this seems to deal with multiple definitions like X + X > Y and 2{X} + Y > Z + X
        for i in t[2][
            0
        ]:  # First tuple member of ReactionEq contains list of (name,stoichcoef)
            if i[0] in self.nDict[ReacID]['Reagents']:
                self.nDict[ReacID]['Reagents'][i[0]] = (
                    self.nDict[ReacID]['Reagents'][i[0]] + i[1]
                )
            else:
                self.nDict[ReacID]['Reagents'][i[0]] = i[
                    1
                ]  # Key for reagent with stoichcoef value
        killList = []
        if __USE_NET_STOICH__:
            # brett: however for the case of X + Y > Y + Z where the sum of the coefficients
            # is zero we can delete the key (Y) out of the reaction list altgether (hopefully!)
            for i in self.nDict[ReacID]['Reagents']:
                if abs(self.nDict[ReacID]['Reagents'][i]) < self.mach_spec_eps2k:
                    killList.append(i)
                    # print self.mach_spec_eps2k, self.nDict[ReacID]['Reagents']
            # print killList, self.nDict[ReacID]['Reagents']
            # brett: and the easiest way of doing this is putting the zero keys in a list
            # and deleting them out of the dictionary
            if len(killList) != 0:
                for i in killList:
                    del self.nDict[ReacID]['Reagents'][i]
            # print killList, self.nDict[ReacID]['Reagents']

        self.nDict[ReacID]['Type'] = t[2][
            1
        ]  # Second tuple member of ReactionEq contains type
        try:  # Save rate equation and create parameter list
            self.nDict[ReacID]['RateEq'] = t[3]
            self.nDict[ReacID]['Params'] = self.ReacParams
            self.ReacParams = []  # Reset global self.ReacParams list
        except:
            self.nDict[ReacID]['RateEq'] = ''
            self.nDict[ReacID]['Params'] = []
            self.AllRateEqsGiven = 0  # Set global flag to false
        self.Show('ReactionLine', t[0])

    def p_reaction_eq(self, t):
        '''ReactionEq : LeftHalfReaction EQUALS RightHalfReaction
                      | LeftHalfReaction IRREV  RightHalfReaction
                      | POOL EQUALS  RightHalfReaction
                      | POOL IRREV  RightHalfReaction
                      | LeftHalfReaction EQUALS POOL
                      | LeftHalfReaction IRREV POOL'''

        ReacType = ''
        if t[2] == '=':
            ReacType = 'Rever'
        elif t[2] == '>':
            ReacType = 'Irrev'

        # Yeah well you know the story ... 4 hrs of trying the "cool, other" way of doing it at work
        # and then solving the original problem in 90 mins with 14 lines of code at home ...
        # oh almost forgot: anonymous source and sink pools now work in PySCeS - brett 20050908
        if t[1] == '$pool':
            t[0] = (t[3], ReacType)
        elif t[3] == '$pool':
            t[0] = (t[1], ReacType)
        else:
            t[0] = (t[1] + t[3], ReacType)
        # print 'reaction_eq',t[0]
        self.Show('ReactionEq', t[0])

    def p_left_half_reaction(self, t):
        ''' LeftHalfReaction : SubstrateTerm
                             | SubstrateTerm PLUS LeftHalfReaction'''
        # Make a list of substrate terms
        t[0] = [t[1]]
        try:
            t[0] += t[3]
        except:
            pass
        self.Show('LeftHalfReaction', t[0])

    def p_right_half_reaction(self, t):
        ''' RightHalfReaction : ProductTerm
                              | ProductTerm PLUS RightHalfReaction'''

        t[0] = [t[1]]

        try:
            t[0] += t[3]
        except:
            pass
        self.Show('RightHalfReaction', t[0])

    def p_substrate_term(self, t):
        '''SubstrateTerm : STOICH_COEF NAME
                         | NAME'''

        # Make tuple of NAME and stoichiometric coefficient
        # (< 0 because substrate)
        try:
            t[0] = (t[2], -float(t[1]))
            if t[2] not in self.Reagents:  # and t[2] != 'self.pool':
                self.Reagents.append(t[2])
        except:
            t[0] = (t[1], -1.0)
            if t[1] not in self.Reagents:  #  and t[1] != 'self.pool':
                self.Reagents.append(t[1])
        self.Show('SubstrateTerm', t[0])

    def p_product_term(self, t):
        '''ProductTerm : STOICH_COEF NAME
                       | NAME'''
        # Make tuple of NAME and stoichiometric coefficient
        # (> 0 because product)
        try:
            t[0] = (t[2], float(t[1]))
            if t[2] not in self.Reagents:  # and t[2] != 'self.pool':
                self.Reagents.append(t[2])
        except:
            t[0] = (t[1], 1.0)
            if t[1] not in self.Reagents:  # and t[1] != 'self.pool':
                self.Reagents.append(t[1])
        self.Show('ProductTerm', t[0])

    def p_rate_eq(self, t):
        '''Expression : Expression PLUS Expression
                      | Expression MINUS Expression
                      | Expression TIMES Expression
                      | Expression DIVIDE Expression
                      | Power
                      | Number
                      | Func'''
        # |UMINUS : add if the
        #  alternative for p_uminus is used

        if len(t.slice) == 4:
            t[0] = t[1] + t[2] + t[3]
        else:
            t[0] = t[1]

    def p_power(self, t):
        '''Power : Expression POWER Expression'''

        t[0] = 'pow(' + t[1] + ',' + t[3] + ')'  # changed to make it DeriVar compatible
        ##  t[0] = t[1] + '**' + t[3]

    def p_uminus(self, t):
        '''Expression : MINUS Expression %prec UMINUS'''
        # Alternative '''UMINUS : MINUS Expression'''

        t[0] = t[1] + t[2]

    def p_number(self, t):
        '''Number : REAL
                  | INT
                  | NAME'''

        # Build list of entities
        ttype = 'param'
        try:
            tx = float(t[1])
            ttype = 'float'
        except ValueError as v:
            pass
            ##  try:
            ##  t[1] = complex(t[1])
            ##  ttype = 'complex'
            ##  except:
            ##  pass
        if ttype in ['complex', 'float']:
            t[1] = str(tx)
        else:
            # if this is not a number add it as a parameter EXCEPT if it is a function's
            # name or numpy.constant
            if (
                (t[1] not in self.ReacParams)
                and (
                    t[1]
                    .replace('numpy.', '')
                    .replace('math.', '')
                    .replace('operator.', '')
                    not in self.MathmlToNumpy_funcs
                )
                and (
                    t[1]
                    .replace('numpy.', '')
                    .replace('math.', '')
                    .replace('operator.', '')
                    not in self.MathmlToNumpy_symb
                )
            ):  # ignore function names and duplications
                self.ReacParams.append(t[1])
        t[0] = t[1]

    # new Core2 based InfixParser code
    def p_function(self, t):
        '''Func : LPAREN ArgList RPAREN
                | NAME LPAREN ArgList RPAREN'''

        # convert root(degree,<expr>) to pow(<expr>, 1/degree)
        if t[1].strip() == 'root':
            t[1] = self.MathmlToNumpy_funcs[t[1]]
            t[3] = '{}, {}'.format(
                t[3][t[3].index(',') + 1 :], 1.0 / float(t[3][: t[3].index(',')]),
            )
            t[0] = t[1] + t[2] + t[3] + t[4]
        elif t[1] in self.MathmlToNumpy_funcs:
            if self.MathmlToNumpy_funcs[t[1]] == None:
                self.SymbolErrors.append(t[1])
                print('\nFunction \"{}\" not supported by PySCeS'.format(t[1]))
                t[0] = 'unknown_function_' + t[1] + t[2] + t[3] + t[4]
            else:
                try:
                    t[0] = self.MathmlToNumpy_funcs[t[1]] + t[2] + t[3] + t[4]
                except Exception as EX:
                    print('Function Parse error 1 (please report!)\n', EX)
            self.ModelUsesNumpyFuncs = True
        # differentiate between bracketed functions and expressions:
        # func( S1 ) and ( S/S05 )
        elif len(t) == 4:
            t[0] = t[1] + t[2] + t[3]
        else:
            # assume some arbitrary function definition
            if (
                t[1][:6] == 'numpy.' or t[1][:5] == 'math.' or t[1][:9] == 'operator.'
            ):  # NEW UNTESTED
                t[0] = t[1] + t[2] + t[3] + t[4]
            else:
                t[0] = t[1] + t[2] + t[3] + t[4]

    # arbitrary number of arguments in an argument list
    # adapted from Andrew Dalke's GardenSnake
    # http://www.dalkescientific.com/writings/diary/GardenSnake.py
    def p_arglist(self, t):
        '''ArgList : Expression
                   | ArgList COMMA Expression'''
        try:
            if len(t) == 2:
                t[0] = t[1]
            elif len(t) == 4:
                t[0] = t[1] + ',' + t[3]
            else:
                pass
        except Exception as EX:
            print('Function ArgList error (please report!)\n', EX)

    ###################################################################
    ###################################################################

    def ParsePSC(self, modelfile, modeldir, modeloutput):
        """
        ParsePSC(modelfile,modeldir,modeloutput)

        Parse the .psc file into a Network Dictionary and associated property arrays

        modelfile: PSC filename
        modeldir: PSC input directory
        modeloutput: working directory for lex/parse temporary files

        """
        # we clear stuff so we have a shiny new instance
        self.nDict = {}  # Dictionary containing all reaction information
        self.InDict = {}
        self.sDict = {}
        self.pDict = {}  # parameter dict
        self.uDict = {
            'substance': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'mole'},
            'volume': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'litre'},
            'time': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'second'},
            'length': {'exponent': 1, 'multiplier': 1.0, 'scale': 0, 'kind': 'metre'},
            'area': {'exponent': 2, 'multiplier': 1.0, 'scale': 0, 'kind': 'metre'},
        }
        self.KeyWords = {
            'Modelname': None,
            'Description': None,
            'Species_In_Conc': None,
            'Output_In_Conc': None,
            'ModelType': None,
        }
        self.compartments = {}
        self.ReactionIDs = []  # List of reaction names
        self.InitStrings = []  # Initialisation strings
        self.species = []  # Variable reagents that occur in reactions
        self.FixedReagents = []  # Fixed reagents
        self.InitParams = []  # Initialised parameters
        self.Names = []  # List of all reagent, parameter and function names
        self.Inits = []  # Initialised entities
        self.Reagents = []  # All reagents found during parsing of reactions
        self.ReacParams = []  # Temporary list of reaction parameters
        self.LexErrors = []  # List of lexing errors
        self.ParseErrors = []
        self.SymbolErrors = []
        self.ForceFunc = []
        self.TimeFunc = []
        self.UserFunc = []
        self.InitFunc = []
        self.Functions = {}
        self.Events = {}
        self.AssignmentRules = {}
        self.UserFuncs = {}
        self.ModelInit = {}
        self.cbm_FluxBounds = {}  # flux bounds in a CB model
        self.cbm_ObjectiveFunctions = {}  # objective functions for a CB model
        self.cbm_UserFluxConstraints = {}  # objective functions for a CB model

        self.ModelUsesNumpyFuncs = False

        # 4 hrs of debugging and these two lines solve the problem .... I'm out of here!
        reload(lex)
        reload(yacc)
        # fixes the obscure reload <--> garbage collection reference overload bug ... who said scripted lang's were
        # easy to use :-) - brett 20040122

        self.AllRateEqsGiven = 1
        self.ParseOK = True
        self.LexOK = True

        # check to see if the last line of the file is an os defined empty line
        self.CheckLastLine(os.path.join(modeldir, modelfile))

        # load model data from file
        Data = open(os.path.join(modeldir, modelfile), 'r')
        self.__Model = Data.read()
        Data.close()

        # try and find a temporary workspace or use ModelOutput
        if 'TMP' in os.environ:
            tempDir = os.environ['TMP']
        elif 'TEMP' in os.environ:
            tempDir = os.environ['TEMP']
        elif os.path.isdir(modeloutput):
            tempDir = modeloutput
        else:
            tempDir = os.getcwd()

        os.chdir(tempDir)

        # fix filenames for intermediary files - brett
        if not modelfile[:-4].isalnum():
            FileL = list(modelfile)
            FileT = ''
            for let in FileL:
                if let.isalnum():
                    FileT += let
            self.debugfile = '_pys' + FileT[:-3] + ".dbg"
            self.tabmodule = '_pys' + FileT[:-3] + "_" + "parsetab"
        else:
            self.debugfile = '_pys' + modelfile[:-4] + ".dbg"
            self.tabmodule = '_pys' + modelfile[:-4] + "_" + "parsetab"

        if self.display_debug:
            print(self.tabmodule)
            print(self.debugfile)
            print('cwd: ', os.getcwd())

        self.__lexer = lex.lex(module=self, debug=self.display_debug)
        self.__lexer.input(self.__Model)
        self.__parser = yacc.yacc(
            module=self,
            debug=self.display_debug,
            debugfile=self.debugfile,
            tabmodule=self.tabmodule,
            outputdir=tempDir,
        )

        while 1:
            tok = self.__lexer.token()
            if not tok:
                break

        while 1:
            p = self.__parser.parse(self.__Model)
            if not p:
                break

        # get to our work directory
        os.chdir(modeloutput)

        # we have the dictionary get rid of this stuff
        del self.__Model, self.__lexer, self.__parser, p

        # Create forcing function code blocks - brett 20050621
        Fstr = ''
        Tstr = ''
        Ustr = ''
        Istr = ''
        for f in self.ForceFunc:
            if os.sys.platform != 'win32' and f[-2] == '\r':
                Fstr = Fstr + f[:-2] + '\n'
            else:
                Fstr += f
        for t in self.TimeFunc:
            if os.sys.platform != 'win32' and t[-2] == '\r':
                Tstr = Tstr + t[:-2] + '\n'
            else:
                Tstr += t
        for u in self.UserFunc:
            if os.sys.platform != 'win32' and u[-2] == '\r':
                Ustr = Ustr + u[:-2] + '\n'
            else:
                Ustr += u
        for i in self.InitFunc:
            if os.sys.platform != 'win32' and i[-2] == '\r':
                Istr = Istr + i[:-2] + '\n'
            else:
                Istr += i

        self.ForceFunc = Fstr
        self.TimeFunc = Tstr
        self.UserFunc = Ustr
        self.InitFunc = Istr
        del Fstr, Tstr, Ustr, Istr

        # Create list of variable species
        for reag in self.Reagents:
            if reag not in self.FixedReagents:
                self.species.append(reag)

        # Warn if no reagents have been fixed
        if self.FixedReagents == []:
            print('Info: No reagents have been fixed')
        else:  # Warn if a fixed reagent does not occur in a reaction equation
            for reag in self.FixedReagents:
                if reag not in self.Reagents:
                    print(
                        'Warning: species "{}" (fixed) does not occur in any reaction'.format(
                            reag.replace('self.', '')
                        )
                    )

        # for the initialised ones
        for s in list(self.sDict.keys()):
            if 'self.' + s in self.FixedReagents + self.Reagents:
                fixed = False
                compartment = None
                isamount = False
                initial = self.sDict[s]['initial']
                name = self.sDict[s]['name']

                if 'self.' + s in self.FixedReagents:
                    fixed = True
                if s in list(self.InDict.keys()):
                    compartment = self.InDict[s]
                self.sDict.update(
                    {
                        s: {
                            'name': name,
                            'initial': initial,
                            'compartment': compartment,
                            'fixed': fixed,
                            'isamount': isamount,
                        }
                    }
                )
            else:
                self.pDict.update({s: self.sDict.pop(s)})

        # for the uninitialised ones
        ##  print self.FixedReagents, self.Reagents, self.Inits
        for s in self.species + self.FixedReagents:
            if s.replace('self.', '') not in self.sDict:
                fixed = False
                compartment = None
                isamount = False
                initial = None
                name = s.replace('self.', '')

                if s in self.FixedReagents:
                    fixed = True
                if name in list(self.InDict.keys()):
                    compartment = self.InDict[name]
                self.sDict.update(
                    {
                        name: {
                            'name': name,
                            'initial': initial,
                            'compartment': compartment,
                            'fixed': fixed,
                            'isamount': isamount,
                        }
                    }
                )

        # Create list of initialised parameters
        # eventually this must be switched over to pDict
        for elem in self.Inits:
            el = elem.replace('self.', '')
            names = [self.sDict[s]['name'] for s in self.sDict]
            if el not in names:
                self.InitParams.append(elem)
            if el in names and self.sDict[el]['fixed'] == True:
                self.InitParams.append(elem)

        # In self.nDict, clean rate equation pameter list of variables that occur in that reaction
        for id in list(self.nDict.keys()):
            # create 'Modifiers' key for each reaction - brett 20050606
            self.nDict[id].update({'Modifiers': []})
            reagcomp = None
            if id in list(self.InDict.keys()):
                reagcomp = self.InDict[id]
            self.nDict[id].update({'compartment': reagcomp})
            for reag in self.species:
                if reag in self.nDict[id]['Params']:
                    # if the reagent is a reaction reagent delete - brett 20050606
                    if reag in self.nDict[id]['Reagents']:
                        self.nDict[id]['Params'].remove(reag)
                    # if not it is a modifier ... add to modifiers and delete - brett 20050606
                    else:
                        # print 'INFO: variable modifier \"' + reag.replace('self.','') + '\" detected in', id
                        self.nDict[id]['Modifiers'].append(reag)
                        self.nDict[id]['Params'].remove(reag)

        # Check whether all parameters have been initialised and if not add to dictionary
        for id in list(self.nDict.keys()):
            cnames = [self.compartments[s]['name'] for s in self.compartments]
            for param in self.nDict[id]['Params']:
                ##  print param
                ##  print self.InitParams
                if (
                    param not in self.InitParams
                    and param.replace('self.', '') not in cnames
                ):
                    ##  print param, self.InitParams, self.pDict
                    print(
                        'Warning: {} parameter "{}"  has not been initialised'.format(
                            id, param.replace('self.', '')
                        )
                    )
                    pname = param.replace('self.', '')
                    self.pDict.update({pname: {'name': pname, 'initial': None}})
                    self.InitParams.append(param)

        # Check whether all variable reagents have been initialised
        for reag in self.species + self.FixedReagents:
            if reag not in self.Inits and reag in self.species:
                print(
                    'Warning: species "{}" has not been initialised'.format(
                        reag.replace('self.', '')
                    )
                )
            if reag not in self.Inits and reag in self.FixedReagents:
                print(
                    'Warning: species "{}" (fixed) has not been initialised'.format(
                        reag.replace('self.', '')
                    )
                )

        # Check that all initialised parameters actually occur in self.Inits
        known = 0
        for param in self.InitParams:
            for id in list(self.nDict.keys()):
                if param in self.nDict[id]['Params']:
                    known = 1
                    break
                else:
                    known = 0
            if not known:
                print(
                    'Info: "{}" has been initialised but does not occur in a rate equation'.format(
                        param.replace('self.', '')
                    )
                )

        # find modifiers for each reaction - brett 20050606
        self.modifiers = []
        for i in self.nDict:
            for j in self.nDict[i]['Params']:
                if j in self.FixedReagents and j not in self.nDict[i]['Reagents']:
                    self.nDict[i]['Modifiers'].append(j)
            self.modifiers.append(
                (i, tuple([j.replace('self.', '') for j in self.nDict[i]['Modifiers']]))
            )

        self.fixed_species = copy.copy(self.FixedReagents)
        self.parameters = copy.copy(self.InitParams)
        self.reactions = copy.copy(self.ReactionIDs)

        for x in range(len(self.fixed_species)):
            self.fixed_species[x] = self.fixed_species[x].replace('self.', '')
        for x in range(len(self.species)):
            self.species[x] = self.species[x].replace('self.', '')
        for x in range(len(self.parameters)):
            self.parameters[x] = self.parameters[x].replace('self.', '')

        if self.display_debug == 1:
            print(self.fixed_species)
            print(self.species)
            print(self.parameters)
            print(self.reactions)

        if self.display_debug == 1:
            print('\nDebugging Part 1')
            print('\nFixedReagents')
            print(self.FixedReagents)
            print('\nInitStrings')
            print(self.InitStrings)
            print('\nInitParams')
            print(self.InitParams)
            print('\nReactionIDs')
            print(self.ReactionIDs)
            print('\nspecies')
            print(self.species)
            print('\nNetworkDict')
            print(self.nDict)
            print('\n\n')

    def KeywordCheck(self, inputarr, bad=[]):
        """
        KeywordCheck(inputarr,bad=[])

        Check a list of names for reserved PySCeS keywords

        Arguments:
        =========
        inputarr: a list of names to check
        bad [default=[]]: a list of illegal keywords (new if not supplied

        """
        for x in inputarr:
            if x.replace('self.', '') in self.ReservedTerms:
                bad.append(x.replace('self.', ''))
                print(x.replace('self.', ''), '\t\tERROR: keyword')
        return bad

    def CheckLastLine(self, name):
        """
        CheckLastLine(name)

        Checks to see if file ends with an empty line ... if not it adds one

        Arguments:
        =========
        name: filename of the PySCeS input file

        """
        F = open(name, 'r+')
        # F.seek(-1,2)
        F.seek(0, os.SEEK_END)
        F.seek(F.tell() - 1, os.SEEK_SET)
        if F.read() != '\n':
            if os.sys.platform == 'win32':
                F.read()
            try:
                F.write('\n')
            except Exception as e:
                print('EOL add error', e)
            print('Adding \\n to input file')
        F.close()


'''
# you REALLY do NOT want to know what this was for! - brett
if gc.isenabled() == 1:
    gc.collect()
    del gc.garbage[:]
    gc.set_debug(0)
    print 'gc debug level = ' + `gc.get_debug()`
'''
